r"""Command line interface entrypoints"""

import numpy as np
import warnings
import random
import typing as tp
import math
import shutil
import sys
import pickle
import multiprocessing as mp
import multiprocessing.shared_memory as shmem
from typing import Annotated
from pathlib import Path

from typer import Typer, Argument, Option, Abort, Context, Exit

from bblean._memory import launch_monitor_rss_daemon
from bblean._timer import Timer
from bblean._config import DEFAULTS, collect_system_specs_and_dump_config, TSNE_SEED
from bblean.utils import _import_bitbirch_variant, batched

app = Typer(
    rich_markup_mode="markdown",
    add_completion=False,
    help=r"""CLI tool for serial or parallel fast clustering of molecular fingerprints
    using the memory-efficient and compute-efficient *O(N)* BitBIRCH algorithm ('Lean'
    version). For more info about the subcommands run `bb <subcommand> --help `.""",
)


def _print_help_banner(ctx: Context, value: bool) -> None:
    if value:
        from bblean._console import get_console

        console = get_console()
        console.print_banner()
        console.print(ctx.get_help())
        raise Exit()


def _validate_output_dir(out_dir: Path, overwrite: bool = False) -> None:
    if out_dir.exists():
        if not out_dir.is_dir():
            raise RuntimeError("Output dir should be a dir")
        if any(out_dir.iterdir()):
            if overwrite:
                shutil.rmtree(out_dir)
            else:
                raise RuntimeError(f"Output dir {out_dir} has files")


# Validate that the naming convention for the input files is correct
def _validate_input_dir(in_dir: Path | str) -> None:
    in_dir = Path(in_dir)
    if not in_dir.is_dir():
        raise RuntimeError(f"Input dir {in_dir} should be a dir")
    fp_files = (f for f in in_dir.glob("*.npy") if not f.stem.endswith(".indices"))
    if not any(fp_files):
        raise RuntimeError(f"Input dir {in_dir} should have *.npy fingerprint files")


@app.callback()
def _main(
    ctx: Context,
    help_: Annotated[
        bool,
        Option(
            "--help/ ",
            "-h",
            is_eager=True,
            help="Show this message and exit.",
            callback=_print_help_banner,
        ),
    ] = False,
) -> None:
    pass


@app.command("compare", rich_help_panel="Analysis", hidden=True)
def _compare(
    clusters_a_path: Annotated[Path, Argument()],
    clusters_b_path: Annotated[Path, Argument()],
    ari: Annotated[
        bool,
        Option("--ari/--no-ari", help="Adjusted Rand index"),
    ] = True,
    ami: Annotated[
        bool,
        Option("--ami/--no-ami", help="Adjusted mutual information (slow)"),
    ] = True,
    top: Annotated[
        int,
        Option("-t", "--top"),
    ] = 30,
    use_first_clustering_indices: Annotated[
        bool,
        Option("--use-first-clustering-indices/--no-use-first-clustering-indices"),
    ] = False,
    json: Annotated[
        bool,
        Option("-j/-J", "--json/--no-json"),
    ] = False,
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
) -> None:
    r"""Compare two clusterings of the same data, using different metrics"""
    import pickle
    import numpy as np

    from sklearn.metrics import adjusted_mutual_info_score, adjusted_rand_score

    from bblean._console import get_console

    console = get_console(silent=not verbose)

    if clusters_a_path.is_dir():
        clusters_a_path = clusters_a_path / "clusters.pkl"

    if clusters_b_path.is_dir():
        clusters_b_path = clusters_b_path / "clusters.pkl"

    with console.status("[italic]Collecting labels...[/italic]", spinner="dots"):
        with open(clusters_a_path, mode="rb") as f:
            clusters = pickle.load(f)
            total = sum(len(c) for c in clusters)
            true_labels = np.empty(total, dtype=np.uint64)
            for i, mol_ids in enumerate(clusters):
                true_labels[mol_ids] = i
            idxs_a = np.concatenate(clusters[:top])

        with open(clusters_b_path, mode="rb") as f:
            clusters = pickle.load(f)
            total = sum(len(c) for c in clusters)
            pred_labels = np.empty(total, dtype=np.uint64)
            for i, mol_ids in enumerate(clusters):
                pred_labels[mol_ids] = i
            idxs_b = np.concatenate(clusters[:top])
    if use_first_clustering_indices:
        idxs = idxs_a
    else:
        idxs = np.unique(np.concatenate((idxs_a, idxs_b)))

    true_labels = true_labels[idxs]
    pred_labels = pred_labels[idxs]
    if json:
        output = {}
        if ami:
            output["ami"] = adjusted_mutual_info_score(true_labels, pred_labels)
        if ari:
            output["ari"] = adjusted_rand_score(true_labels, pred_labels)
        console.print(output, highlight=False)
        sys.exit(0)

    timer = Timer()
    timer.init_timing("total")
    if ami:
        with console.status("[italic]Calc. AMI score...[/italic]", spinner="dots"):
            ami_score = adjusted_mutual_info_score(true_labels, pred_labels)
        console.print(f"Adjusted Mutual Information (AMI): {ami_score:.4f}")

    if ari:
        with console.status("[italic]Calc. ARI score...[/italic]", spinner="dots"):
            ari_score = adjusted_rand_score(true_labels, pred_labels)
        console.print(f"Adjusted Rand Index (ARI): {ari_score:.4f}")
    timer.end_timing("total", console, indent=False)


@app.command("summary", rich_help_panel="Analysis")
def _table_summary(
    clusters_path: Annotated[
        Path,
        Argument(help="Path to the clusters file, or a dir with a clusters.pkl file"),
    ],
    fps_path: Annotated[
        Path | None,
        Option(
            "-f",
            "--fps-path",
            help="Path to fingerprint file, or directory with fingerprint files",
            show_default=False,
        ),
    ] = None,
    min_size: Annotated[
        int,
        Option("--min-size"),
    ] = 0,
    smiles_path: Annotated[
        Path | None,
        Option(
            "-s",
            "--smiles-path",
            show_default=False,
            help="Optional smiles path, if passed a scaffold analysis is performed",
        ),
    ] = None,
    top: Annotated[
        int,
        Option("--top"),
    ] = 20,
    input_is_packed: Annotated[
        bool,
        Option("--packed-input/--unpacked-input", rich_help_panel="Advanced"),
    ] = True,
    scaffold_fp_kind: Annotated[
        str,
        Option("--scaffold-fp-kind"),
    ] = DEFAULTS.fp_kind,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
    metrics: Annotated[
        bool,
        Option(
            "--metrics/--no-metrics",
            help="Calculate clustering indices (Dunn, DBI, CHI)",
        ),
    ] = False,
    chosen_metrics: Annotated[
        str,
        Option(
            "-m",
            "--metrics-choice",
            help=(
                "Chosen metrics. "
                " Comma-separated list including dunn (slow), dbi or chi"
            ),
        ),
    ] = "dunn,dbi,chi",
    metrics_top: Annotated[
        int | None,
        Option("--metrics-top", rich_help_panel="Advanced"),
    ] = 100,
    metrics_min_size: Annotated[
        int,
        Option("--metrics-min-size", hidden=True),
    ] = 1,
    verbose: Annotated[
        bool,
        Option("--verbose/--no-verbose", hidden=True),
    ] = True,
) -> None:
    r"""Summary table of clustering results, together with cluster metrics"""
    from bblean._console import get_console
    from bblean.smiles import load_smiles
    from bblean.analysis import cluster_analysis
    from bblean.utils import _has_files_or_valid_symlinks
    from bblean.metrics import jt_dbi, jt_isim_chi, jt_isim_dunn, _calc_centrals
    from rich.table import Table

    console = get_console(silent=not verbose)
    # Imports may take a bit of time since sklearn is slow, so start the spinner here
    with console.status("[italic]Analyzing clusters...[/italic]", spinner="dots"):
        if clusters_path.is_dir():
            clusters_path = clusters_path / "clusters.pkl"
        with open(clusters_path, mode="rb") as f:
            clusters = pickle.load(f)
        if fps_path is None:
            input_fps_path = clusters_path.parent / "input-fps"
            if input_fps_path.is_dir() and _has_files_or_valid_symlinks(input_fps_path):
                fps_path = input_fps_path
            else:
                msg = (
                    "Could not find input fingerprints. Please use --fps-path."
                    " Summary plot without fingerprints doesn't include isim values"
                )
                warnings.warn(msg)
        if fps_path is None:
            fps_paths = None
        elif fps_path.is_dir():
            fps_paths = sorted(
                f for f in fps_path.glob("*.npy") if not f.stem.endswith(".indices")
            )
        else:
            fps_paths = [fps_path]
        ca = cluster_analysis(
            clusters,
            fps_paths,
            smiles=load_smiles(smiles_path) if smiles_path is not None else (),
            top=top,
            n_features=n_features,
            input_is_packed=input_is_packed,
            min_size=min_size,
        )
        table = Table(title=(f"Top {top} clusters" if top is not None else "Clusters"))
        table.add_column("Size", justify="center")
        table.add_column("% fps", justify="center")
        if ca.has_isims:
            table.add_column("iSIM", justify="center")
        if smiles_path is not None:
            table.add_column("Size/Scaff.", justify="center")
            table.add_column("Num. Scaff.", justify="center")
            table.add_column("Scaff. iSIM", justify="center")
        total_fps = ca.total_fps
        for i in range(ca.clusters_num):
            size = ca.sizes[i]
            percent = size / total_fps * 100
            parts = [f"{size:,}", f"{percent:.2f}"]
            if ca.has_isims:
                parts.append(f"{ca.isims[i]:.3f}")
            table.add_row(*parts)
        console.print(table)
        console.print()
        console.print(f"Total num. fps: {total_fps:,}")
        console.print(f"Total num. clusters: {ca.all_clusters_num:,}")
        singles = ca.all_singletons_num
        singles_percent = singles * 100 / ca.all_clusters_num
        console.print(f"Total num. singletons: {singles:,} ({singles_percent:.2f} %)")
        gt10 = ca.all_clusters_num_with_size_above(10)
        gt10_percent = gt10 * 100 / ca.all_clusters_num
        console.print(
            f"Total num. clusters, size > 10: {gt10:,} ({gt10_percent:.2f} %)"
        )
        gt100 = ca.all_clusters_num_with_size_above(100)
        gt100_percent = gt100 * 100 / ca.all_clusters_num
        console.print(
            f"Total num. clusters, size > 100: {gt100:,} ({gt100_percent:.2f} %)"
        )
        console.print(
            f"num-clusters/num-fps ratio: {ca.all_clusters_num / total_fps:.2f}"
        )
        console.print(f"Mean size: {ca.all_clusters_mean_size:.2f}")
        console.print(f"Max. size: {ca.all_clusters_max_size:,}")
        console.print(f"Q3 (75%) size: {ca.all_clusters_q3:,}")
        console.print(f"Median size: {ca.all_clusters_median_size:,}")
        console.print(f"Q1 (25%) size: {ca.all_clusters_q1:,}")
        console.print(f"Min. size: {ca.all_clusters_min_size:,}")
    if metrics:
        chosen = set(s.lower() for s in chosen_metrics.split(","))
        assert all(s in ["dunn", "chi", "dbi"] for s in chosen)
        # Redo cluster analysis with more *top* clusters
        console.print()
        if metrics_top is None:
            console.print("Clustering metrics:")
        else:
            console.print(f"Clustering metrics considering top {metrics_top} clusters:")
        with console.status("[italic]Reanalyzing clusters...[/italic]", spinner="dots"):
            ca = cluster_analysis(
                clusters,
                fps_paths,
                smiles=(),
                top=metrics_top,
                n_features=n_features,
                input_is_packed=input_is_packed,
                min_size=metrics_min_size,
            )
            clusters = ca.get_top_cluster_fps()
        with console.status("[italic]Calculating centrals...[/italic]", spinner="dots"):
            centrals = _calc_centrals(clusters, kind="centroid")
        if "chi" in chosen:
            chi = jt_isim_chi(clusters, centrals=centrals, verbose=verbose)
            console.print(f"    - CHI index: {chi:.4f} (Higher is better)")
        if "dbi" in chosen:
            dbi = jt_dbi(clusters, centrals=centrals, verbose=verbose)
            console.print(f"    - DBI index: {dbi:.4e} (Lower is better)")
        if "dunn" in chosen:
            dunn = jt_isim_dunn(clusters, verbose=verbose)
            console.print(f"    - Dunn index: {dunn:.4f} (Higher is better)")


@app.command("plot-isim", rich_help_panel="Analysis", hidden=True)
def _plot_isim(
    clusters_path: Annotated[
        Path,
        Argument(help="Path to the clusters file, or a dir with a clusters.pkl file"),
    ],
    fps_path: Annotated[
        Path | None,
        Option(
            "-f",
            "--fps-path",
            help="Path to fingerprint file, or directory with fingerprint files",
            show_default=False,
        ),
    ] = None,
    title: Annotated[
        str | None,
        Option("--title", help="Plot title"),
    ] = None,
    top: Annotated[
        int | None,
        Option("--top"),
    ] = None,
    input_is_packed: Annotated[
        bool,
        Option("--packed-input/--unpacked-input", rich_help_panel="Advanced"),
    ] = True,
    min_size: Annotated[
        int,
        Option("--min-size"),
    ] = 0,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
    save: Annotated[
        bool,
        Option("--save/--no-save"),
    ] = True,
    filename: Annotated[
        str | None,
        Option("--filename"),
    ] = None,
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
    show: Annotated[
        bool,
        Option("--show/--no-show", hidden=True),
    ] = True,
    threshold: Annotated[
        float | None,
        Option("-t", "--threshold"),
    ] = None,
) -> None:
    r"""Population plot of the clustering results"""
    from bblean._console import get_console

    console = get_console(silent=not verbose)
    # Imports may take a bit of time since sklearn is slow, so start the spinner here
    with console.status("[italic]Analyzing clusters...[/italic]", spinner="dots"):
        import json
        from bblean.plotting import (
            _dispatch_visualization,
            isim_dist_plot,
            isim_size_ordered_plot,
        )

        if clusters_path.is_dir():
            clusters_path = clusters_path / "clusters.pkl"
        config_path = clusters_path.parent / "config.json"
        if threshold is None:
            if config_path.exists():
                with open(config_path, mode="rt", encoding="utf-8") as f:
                    threshold = json.load(f)["threshold"]
            else:
                threshold = None
        fig_0, axs_0 = _dispatch_visualization(
            clusters_path,
            "isims-dist",
            isim_dist_plot,
            {"threshold": threshold},
            min_size=min_size,
            top=top,
            n_features=n_features,
            input_is_packed=input_is_packed,
            fps_path=fps_path,
            title=title,
            filename=filename,
            verbose=verbose,
            save=save,
            show=False,
            warn_nan=False,
        )
        _dispatch_visualization(
            clusters_path,
            "isim-ordered",
            isim_size_ordered_plot,
            {"threshold": threshold},
            min_size=min_size,
            top=top,
            n_features=n_features,
            input_is_packed=input_is_packed,
            fps_path=fps_path,
            title=title,
            filename=filename,
            verbose=verbose,
            save=save,
            show=show,
            warn_nan=False,
        )


@app.command("plot-pops", rich_help_panel="Analysis")
def _plot_pops(
    clusters_path: Annotated[
        Path,
        Argument(help="Path to the clusters file, or a dir with a clusters.pkl file"),
    ],
    fps_path: Annotated[
        Path | None,
        Option(
            "-f",
            "--fps-path",
            help="Path to fingerprint file, or directory with fingerprint files",
            show_default=False,
        ),
    ] = None,
    title: Annotated[
        str | None,
        Option("--title", help="Plot title"),
    ] = None,
    top: Annotated[
        int | None,
        Option("--top"),
    ] = None,
    input_is_packed: Annotated[
        bool,
        Option("--packed-input/--unpacked-input", rich_help_panel="Advanced"),
    ] = True,
    min_size: Annotated[
        int,
        Option("--min-size"),
    ] = 0,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
    save: Annotated[
        bool,
        Option("--save/--no-save"),
    ] = True,
    filename: Annotated[
        str | None,
        Option("--filename"),
    ] = None,
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
    show: Annotated[
        bool,
        Option("--show/--no-show", hidden=True),
    ] = True,
) -> None:
    r"""Population plot of the clustering results"""
    from bblean._console import get_console

    console = get_console(silent=not verbose)
    # Imports may take a bit of time since sklearn is slow, so start the spinner here
    with console.status("[italic]Analyzing clusters...[/italic]", spinner="dots"):
        from bblean.plotting import _dispatch_visualization, pops_plot

        _dispatch_visualization(
            clusters_path,
            "pops",
            pops_plot,
            {},
            min_size=min_size,
            top=top,
            n_features=n_features,
            input_is_packed=input_is_packed,
            fps_path=fps_path,
            title=title,
            filename=filename,
            verbose=verbose,
            save=save,
            show=show,
        )


@app.command("plot-umap", rich_help_panel="Analysis")
def _plot_umap(
    clusters_path: Annotated[
        Path,
        Argument(help="Path to the clusters file, or a dir with a clusters.pkl file"),
    ],
    fps_path: Annotated[
        Path | None,
        Option(
            "-f",
            "--fps-path",
            help="Path to fingerprint file, or directory with fingerprint files",
            show_default=False,
        ),
    ] = None,
    title: Annotated[
        str | None,
        Option("--title", help="Plot title"),
    ] = None,
    save: Annotated[
        bool,
        Option("--save/--no-save"),
    ] = True,
    top: Annotated[
        int,
        Option("--top"),
    ] = 20,
    input_is_packed: Annotated[
        bool,
        Option("--packed-input/--unpacked-input", rich_help_panel="Advanced"),
    ] = True,
    scaling: Annotated[
        str,
        Option("--scaling", rich_help_panel="Advanced"),
    ] = "normalize",
    min_size: Annotated[
        int,
        Option("--min-size"),
    ] = 0,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
    filename: Annotated[
        str | None,
        Option("--filename"),
    ] = None,
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
    show: Annotated[
        bool,
        Option("--show/--no-show", hidden=True),
    ] = True,
    deterministic: Annotated[
        bool,
        Option("--deterministic/--no-deterministic"),
    ] = False,
    n_neighbors: Annotated[
        int,
        Option("-n", "--neighbors"),
    ] = 15,
    min_dist: Annotated[
        float,
        Option("-d", "--min-dist"),
    ] = 0.5,
    metric: Annotated[
        str,
        Option("--metric"),
    ] = "euclidean",
    densmap: Annotated[
        bool,
        Option("--densmap/--no-densmap"),
    ] = False,
    workers: Annotated[
        int | None,
        Option(
            "-w",
            "--workers",
            help="Num. cores to use for parallel processing",
            rich_help_panel="Advanced",
        ),
    ] = None,
) -> None:
    r"""UMAP visualization of the clustering results"""
    from bblean._console import get_console

    console = get_console(silent=not verbose)
    # Imports may take a bit of time since sklearn is slow, so start the spinner here
    with console.status("[italic]Analyzing clusters...[/italic]", spinner="dots"):
        from bblean.plotting import _dispatch_visualization, umap_plot

        kwargs = dict(
            metric=metric,
            densmap=densmap,
            deterministic=deterministic,
            n_neighbors=n_neighbors,
            workers=workers,
            min_dist=min_dist,
        )
        _dispatch_visualization(
            clusters_path,
            "umap",
            umap_plot,
            kwargs,
            min_size=min_size,
            top=top,
            n_features=n_features,
            input_is_packed=input_is_packed,
            fps_path=fps_path,
            title=title,
            filename=filename,
            verbose=verbose,
            save=save,
            show=show,
        )


@app.command("plot-pca", rich_help_panel="Analysis")
def _plot_pca(
    clusters_path: Annotated[
        Path,
        Argument(help="Path to the clusters file, or a dir with a clusters.pkl file"),
    ],
    fps_path: Annotated[
        Path | None,
        Option(
            "-f",
            "--fps-path",
            help="Path to fingerprint file, or directory with fingerprint files",
            show_default=False,
        ),
    ] = None,
    title: Annotated[
        str | None,
        Option("--title", help="Plot title"),
    ] = None,
    top: Annotated[
        int,
        Option("--top"),
    ] = 20,
    min_size: Annotated[
        int,
        Option("--min-size"),
    ] = 0,
    input_is_packed: Annotated[
        bool,
        Option("--packed-input/--unpacked-input", rich_help_panel="Advanced"),
    ] = True,
    scaling: Annotated[
        str,
        Option("--scaling", rich_help_panel="Advanced"),
    ] = "normalize",
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
    show: Annotated[
        bool,
        Option("--show/--no-show", hidden=True),
    ] = True,
    whiten: Annotated[
        bool,
        Option("--whiten/--no-whiten"),
    ] = False,
    save: Annotated[
        bool,
        Option("--save/--no-save"),
    ] = True,
    filename: Annotated[
        str | None,
        Option("--filename"),
    ] = None,
) -> None:
    r"""PCA visualization of the clustering results"""
    from bblean._console import get_console

    console = get_console(silent=not verbose)
    # Imports may take a bit of time since sklearn is slow, so start the spinner here
    with console.status("[italic]Analyzing clusters...[/italic]", spinner="dots"):
        from bblean.plotting import _dispatch_visualization, pca_plot

        _dispatch_visualization(
            clusters_path,
            "pca",
            pca_plot,
            {"whiten": whiten},
            min_size=min_size,
            top=top,
            n_features=n_features,
            input_is_packed=input_is_packed,
            fps_path=fps_path,
            title=title,
            filename=filename,
            verbose=verbose,
            save=save,
            show=show,
        )


@app.command("plot-tsne", rich_help_panel="Analysis")
def _plot_tsne(
    clusters_path: Annotated[
        Path,
        Argument(help="Path to the clusters file, or a dir with a clusters.pkl file"),
    ],
    fps_path: Annotated[
        Path | None,
        Option(
            "-f",
            "--fps-path",
            help="Path to fingerprint file, or directory with fingerprint files",
            show_default=False,
        ),
    ] = None,
    title: Annotated[
        str | None,
        Option("--title", help="Plot title"),
    ] = None,
    save: Annotated[
        bool,
        Option("--save/--no-save"),
    ] = True,
    min_size: Annotated[
        int,
        Option("--min-size"),
    ] = 0,
    filename: Annotated[
        str | None,
        Option("--filename"),
    ] = None,
    exaggeration: Annotated[
        float | None,
        Option("-e", "--exaggeration", rich_help_panel="Advanced"),
    ] = None,
    seed: Annotated[
        int | None,
        Option(
            "-s",
            "--seed",
            help=(
                "Seed for the rng, fixed value by default, for reproducibility."
                " Pass -1 to randomize"
            ),
            show_default=False,
            rich_help_panel="Advanced",
        ),
    ] = TSNE_SEED,
    top: Annotated[
        int,
        Option("--top"),
    ] = 20,
    metric: Annotated[
        str,
        Option("--metric", help="Metric to use in the t-SNE source space"),
    ] = "euclidean",
    dof: Annotated[
        float,
        Option("-d", "--dof", rich_help_panel="Advanced"),
    ] = 1.0,
    perplexity: Annotated[
        int,
        Option(help="t-SNE perplexity", rich_help_panel="Advanced"),
    ] = 30,
    input_is_packed: Annotated[
        bool,
        Option("--packed-input/--unpacked-input", rich_help_panel="Advanced"),
    ] = True,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
    scaling: Annotated[
        str,
        Option("--scaling", rich_help_panel="Advanced"),
    ] = "normalize",
    do_pca_init: Annotated[
        bool,
        Option(
            "--pca-init/--no-pca-init",
            rich_help_panel="Advanced",
            help="Use PCA for initialization",
        ),
    ] = True,
    pca_reduce: Annotated[
        int | None,
        Option(
            "-p",
            "--pca-reduce",
            rich_help_panel="Advanced",
            help=(
                "Reduce fingerprint dimensionality to N components using PCA."
                " A value of 50 or more maintains cluster structure in general"
            ),
        ),
    ] = None,
    workers: Annotated[
        int | None,
        Option(
            "-w",
            "--workers",
            help="Num. cores to use for parallel processing",
            rich_help_panel="Advanced",
        ),
    ] = None,
    multiscale: Annotated[
        bool,
        Option(
            "-m/-M",
            "--multiscale/--no-multiscale",
            rich_help_panel="Advanced",
            help="Use multiscale perplexities (WARNING: Can be very slow!)",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
    show: Annotated[
        bool,
        Option("--show/--no-show", hidden=True),
    ] = True,
) -> None:
    r"""t-SNE visualization of the clustering results"""
    from bblean._console import get_console

    console = get_console(silent=not verbose)
    # Imports may take a bit of time since sklearn is slow, so start the spinner here
    with console.status("[italic]Analyzing clusters...[/italic]", spinner="dots"):
        from bblean.plotting import _dispatch_visualization, tsne_plot

        kwargs = dict(
            metric=metric,
            seed=seed,
            perplexity=perplexity,
            exaggeration=exaggeration,
            dof=dof,
            workers=workers,
            scaling=scaling,
            do_pca_init=do_pca_init,
            multiscale=multiscale,
            pca_reduce=pca_reduce,
        )
        _dispatch_visualization(
            clusters_path,
            "tsne",
            tsne_plot,
            kwargs,
            min_size=min_size,
            top=top,
            n_features=n_features,
            input_is_packed=input_is_packed,
            fps_path=fps_path,
            title=title,
            filename=filename,
            verbose=verbose,
            save=save,
            show=show,
        )


@app.command("plot-summary", rich_help_panel="Analysis")
def _plot_summary(
    clusters_path: Annotated[
        Path,
        Argument(help="Path to the clusters file, or a dir with a clusters.pkl file"),
    ],
    fps_path: Annotated[
        Path | None,
        Option(
            "-f",
            "--fps-path",
            help="Path to fingerprint file, or directory with fingerprint files",
            show_default=False,
        ),
    ] = None,
    save: Annotated[
        bool,
        Option("--save/--no-save"),
    ] = True,
    ylim: Annotated[
        int | None,
        Option("--ylim"),
    ] = None,
    min_size: Annotated[
        int,
        Option("--min-size"),
    ] = 0,
    smiles_path: Annotated[
        Path | None,
        Option(
            "-s",
            "--smiles-path",
            show_default=False,
            help="Optional smiles path, if passed a scaffold analysis is performed",
        ),
    ] = None,
    title: Annotated[
        str | None,
        Option("--title"),
    ] = None,
    filename: Annotated[
        str | None,
        Option("--filename"),
    ] = None,
    top: Annotated[
        int,
        Option("--top"),
    ] = 20,
    input_is_packed: Annotated[
        bool,
        Option("--packed-input/--unpacked-input", rich_help_panel="Advanced"),
    ] = True,
    scaffold_fp_kind: Annotated[
        str,
        Option("--scaffold-fp-kind"),
    ] = DEFAULTS.fp_kind,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
    annotate: Annotated[
        bool,
        Option(
            "--annotate/--no-annotate",
            help="Display scaffold and fingerprint number in each cluster",
        ),
    ] = True,
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
    show: Annotated[
        bool,
        Option("--show/--no-show", hidden=True),
    ] = True,
) -> None:
    r"""Summary plot of the clustering results"""
    from bblean._console import get_console

    console = get_console(silent=not verbose)
    # Imports may take a bit of time since sklearn is slow, so start the spinner here
    with console.status("[italic]Analyzing clusters...[/italic]", spinner="dots"):
        from bblean.plotting import _dispatch_visualization, summary_plot
        from bblean.smiles import load_smiles

        _dispatch_visualization(
            clusters_path,
            "summary",
            summary_plot,
            {"annotate": annotate, "counts_ylim": ylim},
            smiles=load_smiles(smiles_path) if smiles_path is not None else (),
            min_size=min_size,
            top=top,
            n_features=n_features,
            input_is_packed=input_is_packed,
            fps_path=fps_path,
            title=title,
            filename=filename,
            verbose=verbose,
            save=save,
            show=show,
        )


@app.command("thresh")
def _guess_threshold(
    ctx: Context,
    input_: Annotated[
        Path,
        Argument(help="`*.npy` file with fingerprints"),
    ],
    factor: Annotated[
        float,
        Option("-f", "--factor"),
    ] = 3.0,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " It must be provided for packed inputs *if it is not a multiple of 8*."
            " For typical fingerprint sizes (e.g. 2048, 1024), it is not required",
            rich_help_panel="Advanced",
        ),
    ] = None,
    input_is_packed: Annotated[
        bool,
        Option(
            "--packed-input/--unpacked-input",
            help="Toggle whether the input consists on packed or unpacked fingerprints",
            rich_help_panel="Advanced",
        ),
    ] = True,
    max_samples: Annotated[
        int,
        Option("-m", "--max-samples"),
    ] = 1_000_000,
) -> None:
    r"""Estimate the optimal BitBirch threshold for a fingerprints file"""
    from bblean.bitbirch import guess_threshold
    from bblean._console import get_console

    console = get_console()
    fps = np.load(input_)
    thresh, mean, std = guess_threshold(
        fps, input_is_packed, n_features, max_samples, factor, return_mean_std=True
    )
    console.print(f"Estimated average similarity: {mean:.4f}")
    console.print(f"Estimated similarity deviation: {std:.4f}")
    console.print(f"Estimated optimal threshold: {thresh:.4f}")


@app.command("run")
def _run(
    ctx: Context,
    input_: Annotated[
        Path | None,
        Argument(help="`*.npy` file with fingerprints, or dir with `*.npy` files"),
    ] = None,
    out_dir: Annotated[
        Path | None,
        Option(
            "-o",
            "--out-dir",
            help="Dir to dump the output files",
        ),
    ] = None,
    overwrite: Annotated[bool, Option(help="Allow overwriting output files")] = False,
    branching_factor: Annotated[
        int,
        Option(
            "--branching",
            "-b",
            help="BitBIRCH branching factor (all rounds). Usually 254 is"
            " optimal. Set above 254 for slightly less RAM (at the cost of some perf.)",
        ),
    ] = DEFAULTS.branching_factor,
    threshold: Annotated[
        float,
        Option("--threshold", "-t", help="Threshold for merge criterion"),
    ] = DEFAULTS.threshold,
    refine_threshold_change: Annotated[
        float,
        Option(
            "--refine-threshold-change",
            help="Modify threshold for refinement criterion, can be negative",
            hidden=True,
        ),
    ] = DEFAULTS.refine_threshold_change,
    save_tree: Annotated[
        bool,
        Option("--save-tree/--no-save-tree", rich_help_panel="Advanced"),
    ] = False,
    save_centroids: Annotated[
        bool,
        Option("--save-centroids/--no-save-centroids", rich_help_panel="Advanced"),
    ] = True,
    merge_criterion: Annotated[
        str,
        Option("--set-merge", "-m", help="Merge criterion for initial clustsering"),
    ] = DEFAULTS.merge_criterion,
    refine_merge_criterion: Annotated[
        str,
        Option("--set-refine-merge", help="Merge criterion for refinement clustsering"),
    ] = DEFAULTS.refine_merge_criterion,
    tolerance: Annotated[
        float,
        Option(help="BitBIRCH tolerance. For refinement and --set-merge 'tolerance'"),
    ] = DEFAULTS.tolerance,
    refine_num: Annotated[
        int,
        Option(
            "--refine-num",
            help=(
                "Num. of largest clusters to refine."
                " 1 for standard refinement, 0 is the default (no refinement)"
            ),
            hidden=True,
        ),
    ] = 0,
    refine_rounds: Annotated[
        int | None,
        Option(
            "--refine-rounds",
            help=("Num. of refinement rounds. "),
        ),
    ] = None,
    recluster_rounds: Annotated[
        int,
        Option(
            "--recluster-rounds",
            help=("Num. of reclustering rounds. "),
        ),
    ] = 0,
    recluster_shuffle: Annotated[
        bool,
        Option("--recluster-shuffle/--no-recluster-shuffle", hidden=True),
    ] = False,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " It must be provided for packed inputs *if it is not a multiple of 8*."
            " For typical fingerprint sizes (e.g. 2048, 1024), it is not required",
            rich_help_panel="Advanced",
        ),
    ] = None,
    input_is_packed: Annotated[
        bool,
        Option(
            "--packed-input/--unpacked-input",
            help="Toggle whether the input consists on packed or unpacked fingerprints",
            rich_help_panel="Advanced",
        ),
    ] = True,
    # Debug options
    monitor_rss: Annotated[
        bool,
        Option(
            "--monitor-mem/--no-monitor-mem",
            help="Monitor RAM used by all processes",
            rich_help_panel="Advanced",
        ),
    ] = True,
    monitor_rss_interval_s: Annotated[
        float,
        Option(
            "--monitor-mem-seconds",
            help="Interval in seconds for RAM monitoring",
            rich_help_panel="Debug",
            hidden=True,
        ),
    ] = 1.0,
    max_fps: Annotated[
        int | None,
        Option(
            help="Max. num of fingerprints to read from each file",
            rich_help_panel="Debug",
            hidden=True,
        ),
    ] = None,
    variant: Annotated[
        str,
        Option(
            "--bb-variant",
            help="Use different bitbirch variants, *only for debugging*.",
            hidden=True,
        ),
    ] = "lean",
    copy_inputs: Annotated[
        bool,
        Option(
            "--copy/--no-copy",
            rich_help_panel="Advanced",
            help="Copy the input files instead of symlink",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
) -> None:
    r"""Run standard, serial BitBIRCH clustering over `*.npy` fingerprint files"""
    # TODO: Remove code duplication with multiround
    from bblean._console import get_console
    from bblean.fingerprints import _get_fps_file_num

    console = get_console(silent=not verbose)
    if variant == "int64" and input_is_packed:
        raise ValueError("Packed inputs are not supported for the int64 variant")
    if refine_rounds is None:
        refine_rounds = 1 if refine_num > 0 else 0
    if refine_rounds > 0 and refine_num == 0:
        refine_num = 1
    ctx.params["refine_rounds"] = refine_rounds
    ctx.params["refine_num"] = refine_num

    BitBirch, set_merge = _import_bitbirch_variant(variant)

    # NOTE: Files are sorted according to name
    if input_ is None:
        input_ = Path.cwd() / "bb_inputs"
        input_.mkdir(exist_ok=True)
        input_files = sorted(
            f for f in input_.glob("*.npy") if not f.stem.endswith(".indices")
        )
        _validate_input_dir(input_)
    elif input_.is_dir():
        input_files = sorted(
            f for f in input_.glob("*.npy") if not f.stem.endswith(".indices")
        )
        _validate_input_dir(input_)
    else:
        input_files = [input_]
    ctx.params.pop("input_")
    ctx.params["input_files"] = [str(p.resolve()) for p in input_files]
    ctx.params["num_fps_present"] = [_get_fps_file_num(p) for p in input_files]
    if max_fps is not None:
        ctx.params["num_fps_loaded"] = [
            min(n, max_fps) for n in ctx.params["num_fps_present"]
        ]
    else:
        ctx.params["num_fps_loaded"] = ctx.params["num_fps_present"]
    unique_id = format(random.getrandbits(32), "08x")
    if out_dir is None:
        out_dir = Path.cwd() / "bb_run_outputs" / unique_id
    out_dir.mkdir(exist_ok=True, parents=True)
    _validate_output_dir(out_dir, overwrite)
    ctx.params["out_dir"] = str(out_dir.resolve())

    console.print_banner()
    console.print()
    console.print_config(ctx.params)

    # Optinally start a separate process that tracks RAM usage
    if monitor_rss:
        launch_monitor_rss_daemon(out_dir / "monitor-rss.csv", monitor_rss_interval_s)

    timer = Timer()
    timer.init_timing("total")
    if "lean" not in variant:
        set_merge(merge_criterion, tolerance=tolerance)
        tree = BitBirch(branching_factor=branching_factor, threshold=threshold)
    else:
        tree = BitBirch(
            branching_factor=branching_factor,
            threshold=threshold,
            merge_criterion=merge_criterion,
            tolerance=tolerance,
        )
    with console.status("[italic]BitBirching...[/italic]", spinner="dots"):
        for file in input_files:
            # Fitting a file uses mmap internally, and releases memory in a smart way
            tree.fit(
                file,
                n_features=n_features,
                input_is_packed=input_is_packed,
                max_fps=max_fps,
            )
    if recluster_rounds != 0 or refine_rounds != 0:
        tree.set_merge(
            refine_merge_criterion,
            tolerance=tolerance,
            threshold=threshold + refine_threshold_change,
        )
        for r in range(refine_rounds):
            msg = (
                f"[italic]Refinement, round {r + 1}"
                f" (will split {refine_num} largest clusters)...[/italic]"
            )
            with console.status(msg, spinner="dots"):
                tree.refine_inplace(
                    input_files,
                    input_is_packed=input_is_packed,
                    n_largest=refine_num,
                )
        for r in range(recluster_rounds):
            msg = f"[italic]Reclustering, round {r + 1}...[/italic]"
            with console.status(msg, spinner="dots"):
                tree.recluster_inplace(shuffle=recluster_shuffle)

    timer.end_timing("total", console, indent=False)
    console.print_peak_mem(out_dir, indent=False)
    if save_tree:
        if variant != "lean":
            console.print("Can't save tree for non-lean variants", style="red")
        else:
            # TODO: Find alternative solution
            tree.save(out_dir / "bitbirch.pkl")
    if variant == "lean":
        tree.delete_internal_nodes()
    # Dump outputs (peak memory, timings, config, cluster ids)
    if save_centroids:
        if variant != "lean":
            console.print("Can't save centroids for non-lean variants", style="red")
            with open(out_dir / "clusters.pkl", mode="wb") as f:
                pickle.dump(tree.get_cluster_mol_ids(), f)
        else:
            output = tree.get_centroids_mol_ids()
            with open(out_dir / "clusters.pkl", mode="wb") as f:
                pickle.dump(output["mol_ids"], f)
            with open(out_dir / "cluster-centroids-packed.pkl", mode="wb") as f:
                pickle.dump(output["centroids"], f)
    else:
        with open(out_dir / "clusters.pkl", mode="wb") as f:
            pickle.dump(tree.get_cluster_mol_ids(), f)

    collect_system_specs_and_dump_config(ctx.params)
    timer.dump(out_dir / "timings.json")

    # Symlink or copy fingerprint files
    input_fps_dir = (out_dir / "input-fps").resolve()
    input_fps_dir.mkdir()
    if copy_inputs:
        for file in input_files:
            shutil.copy(file, input_fps_dir / file.name)
    else:
        for file in input_files:
            (input_fps_dir / file.name).symlink_to(file.resolve())


# TODO: Currently sometimes after a round is triggered *more* files are output, since
# the files are divided *both* by uint8/uint16 and the batch idx. I believe this is not
# ideal
@app.command("multiround")
def _multiround(
    ctx: Context,
    in_dir: Annotated[
        Path | None,
        Argument(help="Directory with input `*.npy` files with packed fingerprints"),
    ] = None,
    out_dir: Annotated[
        Path | None,
        Option("-o", "--out-dir", help="Dir for output files"),
    ] = None,
    overwrite: Annotated[bool, Option(help="Allow overwriting output files")] = False,
    num_initial_processes: Annotated[
        int, Option("--ps", "--processes", help="Num. processes for first round")
    ] = 10,
    num_midsection_processes: Annotated[
        int | None,
        Option(
            "--mid-ps",
            "--mid-processes",
            help="Num. processes for middle section rounds."
            " These are memory intensive,"
            " you may want to use 50%-30% of --ps."
            " Default is same as --ps",
        ),
    ] = None,
    branching_factor: Annotated[
        int,
        Option(
            "--branching",
            "-b",
            help="BitBIRCH branching factor (all rounds). Usually 254 is"
            " optimal. Set above 254 for slightly less RAM (at the cost of some perf.)",
        ),
    ] = DEFAULTS.branching_factor,
    threshold: Annotated[
        float,
        Option("--threshold", "-t", help="Thresh for merge criterion (initial step)"),
    ] = DEFAULTS.threshold,
    initial_merge_criterion: Annotated[
        str,
        Option(
            "--set-merge",
            "-m",
            help="Initial merge criterion for round 1. ('diameter' recommended)",
        ),
    ] = DEFAULTS.merge_criterion,
    save_tree: Annotated[
        bool,
        Option("--save-tree/--no-save-tree", rich_help_panel="Advanced"),
    ] = False,
    save_centroids: Annotated[
        bool,
        Option("--save-centroids/--no-save-centroids", rich_help_panel="Advanced"),
    ] = True,
    mid_threshold_change: Annotated[
        float,
        Option(
            "--mid-threshold-change",
            help="Modify threshold for refinement",
            rich_help_panel="Advanced",
        ),
    ] = DEFAULTS.refine_threshold_change,
    mid_merge_criterion: Annotated[
        str,
        Option(
            "--set-mid-merge",
            help="Merge criterion for mid rounds ('tolerance-diameter' recommended)",
        ),
    ] = DEFAULTS.refine_merge_criterion,
    tolerance: Annotated[
        float,
        Option(
            help="Tolerance value for all steps that use the 'tolerance' criterion"
            " (by default all except initial round)",
        ),
    ] = DEFAULTS.tolerance,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
    input_is_packed: Annotated[
        bool,
        Option(
            "--packed-input/--unpacked-input",
            help="Toggle whether the input consists on packed or unpacked fingerprints",
            rich_help_panel="Advanced",
        ),
    ] = True,
    # Advanced options
    num_midsection_rounds: Annotated[
        int,
        Option(
            "--num-mid-rounds",
            help="Number of midsection rounds to perform",
        ),
    ] = 1,
    split_largest_after_midsection: Annotated[
        bool,
        Option(
            "--split-after-mid/--no-split-after-mid",
            help=(
                "Split largest cluster after each midsection round"
                " (to be refined by the next round)"
            ),
            rich_help_panel="Advanced",
        ),
    ] = False,
    refinement_before_midsection: Annotated[
        str,
        Option(
            "--initial-refine",
            help=(
                "Run a *full* refinement step after the initial clustering round,"
                " only *split* largest cluster, or do *none*."
            ),
            rich_help_panel="Advanced",
        ),
    ] = "full",
    max_tasks_per_process: Annotated[
        int, Option(help="Max tasks per process", rich_help_panel="Advanced")
    ] = 1,
    fork: Annotated[
        bool,
        Option(
            help="In linux, force the 'fork' multiprocessing start method",
            rich_help_panel="Advanced",
        ),
    ] = False,
    bin_size: Annotated[
        int,
        Option(help="Bin size for chunking during Round 2", rich_help_panel="Advanced"),
    ] = 10,
    # Debug options
    variant: Annotated[
        str,
        Option(
            "--bb-variant",
            help="Use different bitbirch variants, *only for debugging*.",
            hidden=True,
        ),
    ] = "lean",
    monitor_rss: Annotated[
        bool,
        Option(
            "--monitor-mem/--no-monitor-mem",
            help="Monitor RAM used by all processes",
            rich_help_panel="Advanced",
        ),
    ] = True,
    monitor_rss_interval_s: Annotated[
        float,
        Option(
            "--monitor-mem-seconds",
            help="Interval in seconds for RAM monitoring",
            rich_help_panel="Debug",
            hidden=True,
        ),
    ] = 1.0,
    max_fps: Annotated[
        int | None,
        Option(
            help="Max num. of fps to load from each input file",
            rich_help_panel="Debug",
            hidden=True,
        ),
    ] = None,
    max_files: Annotated[
        int | None,
        Option(help="Max num. files to read", rich_help_panel="Debug", hidden=True),
    ] = None,
    copy_inputs: Annotated[
        bool,
        Option(
            "--copy/--no-copy",
            rich_help_panel="Advanced",
            help="Copy the input files instead of symlink",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
    cleanup: Annotated[
        bool,
        Option("--cleanup/--no-cleanup", hidden=True),
    ] = True,
) -> None:
    r"""Run multi-round BitBIRCH clustering, optionally parallelize over `*.npy` files"""  # noqa:E501
    from bblean._console import get_console
    from bblean.multiround import run_multiround_bitbirch
    from bblean.fingerprints import _get_fps_file_num

    console = get_console(silent=not verbose)

    # Set multiprocessing start method
    if fork and not sys.platform == "linux":
        console.print("'fork' is only available on Linux", style="red")
        raise Abort()
    if sys.platform == "linux":
        mp_context = mp.get_context("fork" if fork else "forkserver")
    else:
        mp_context = mp.get_context()

    # Collect inputs:
    # If not passed, input dir is bb_inputs/
    if in_dir is None:
        in_dir = Path.cwd() / "bb_inputs"
    _validate_input_dir(in_dir)
    # All files in the input dir with *.npy suffix are considered input files
    input_files = sorted(
        f for f in in_dir.glob("*.npy") if not f.stem.endswith(".indices")
    )[:max_files]
    ctx.params["input_files"] = [str(p.resolve()) for p in input_files]
    ctx.params["num_fps"] = [_get_fps_file_num(p) for p in input_files]
    if max_fps is not None:
        ctx.params["num_fps_loaded"] = [min(n, max_fps) for n in ctx.params["num_fps"]]
    else:
        ctx.params["num_fps_loaded"] = ctx.params["num_fps"]

    # Set up outputs:
    # If not passed, output dir is constructed as bb_multiround_outputs/<unique-id>/
    unique_id = format(random.getrandbits(32), "08x")
    if out_dir is None:
        out_dir = Path.cwd() / "bb_multiround_outputs" / unique_id
    out_dir.mkdir(exist_ok=True, parents=True)
    _validate_output_dir(out_dir, overwrite)
    ctx.params["out_dir"] = str(out_dir.resolve())

    console.print_banner()
    console.print()
    console.print_multiround_config(ctx.params, mp_context)

    # Optinally start a separate process that tracks RAM usage
    if monitor_rss:
        launch_monitor_rss_daemon(out_dir / "monitor-rss.csv", monitor_rss_interval_s)

    timer = run_multiround_bitbirch(
        input_files=input_files,
        n_features=n_features,
        input_is_packed=input_is_packed,
        out_dir=out_dir,
        initial_merge_criterion=initial_merge_criterion,
        midsection_merge_criterion=mid_merge_criterion,
        num_initial_processes=num_initial_processes,
        num_midsection_processes=num_midsection_processes,
        branching_factor=branching_factor,
        threshold=threshold,
        midsection_threshold_change=mid_threshold_change,
        tolerance=tolerance,
        # Advanced
        save_tree=save_tree,
        save_centroids=save_centroids,
        bin_size=bin_size,
        max_tasks_per_process=max_tasks_per_process,
        refinement_before_midsection=refinement_before_midsection,
        num_midsection_rounds=num_midsection_rounds,
        split_largest_after_each_midsection_round=split_largest_after_midsection,
        # Debug
        max_fps=max_fps,
        verbose=verbose,
        mp_context=mp_context,
        cleanup=cleanup,
    )
    timer.dump(out_dir / "timings.json")
    # TODO: Also dump peak-rss.json
    collect_system_specs_and_dump_config(ctx.params)

    # Symlink or copy fingerprint files
    input_fps_dir = (out_dir / "input-fps").resolve()
    input_fps_dir.mkdir()
    if copy_inputs:
        for file in input_files:
            shutil.copy(file, input_fps_dir / file.name)
    else:
        for file in input_files:
            (input_fps_dir / file.name).symlink_to(file.resolve())


@app.command("fps-info", rich_help_panel="Fingerprints")
def _fps_info(
    fp_paths: Annotated[
        list[Path] | None,
        Argument(show_default=False, help="Paths to *.smi files with smiles"),
    ] = None,
) -> None:
    """Show info about a `*.npy` fingerprint file, or a dir with `*.npy` files"""
    from bblean._console import get_console
    from bblean.fingerprints import _print_fps_file_info

    console = get_console()
    if fp_paths is None:
        fp_paths = [Path.cwd()]

    for path in fp_paths:
        if path.is_dir():
            for file in path.glob("*.npy"):
                if file.stem.endswith(".indices"):
                    continue
                _print_fps_file_info(file, console)
        elif path.suffix == ".npy":
            _print_fps_file_info(path, console)


@app.command("fps-from-smiles", rich_help_panel="Fingerprints")
def _fps_from_smiles(
    smiles_paths: Annotated[
        list[Path] | None,
        Argument(show_default=False, help="Paths to *.smi files with smiles"),
    ] = None,
    out_dir: Annotated[
        Path | None,
        Option("-o", "--out-dir", show_default=False),
    ] = None,
    out_name: Annotated[
        str | None,
        Option("--name", help="Base name of output file"),
    ] = None,
    kind: Annotated[
        str,
        Option("-k", "--kind"),
    ] = DEFAULTS.fp_kind,
    fp_size: Annotated[
        int,
        Option("--n-features", help="Num. features of the generated fingerprints"),
    ] = DEFAULTS.n_features,
    parts: Annotated[
        int | None,
        Option(
            "-n", "--num-parts", help="Split the created file into this number of parts"
        ),
    ] = None,
    max_fps_per_file: Annotated[
        int | None,
        Option(
            "-m",
            "--max-fps-per-file",
            help="Max. number of fps per file. Mutually exclusive with --num-parts",
            show_default=False,
        ),
    ] = None,
    pack: Annotated[
        bool,
        Option(
            "-p/-P",
            "--pack/--no-pack",
            help="Pack bits in last dimension of fingerprints",
            rich_help_panel="Advanced",
        ),
    ] = True,
    dtype: Annotated[
        str,
        Option(
            "-d",
            "--dtype",
            help="NumPy dtype for the generated fingerprints",
            rich_help_panel="Advanced",
        ),
    ] = "uint8",
    verbose: Annotated[
        bool,
        Option("-v/-V", "--verbose/--no-verbose"),
    ] = True,
    num_ps: Annotated[
        int | None,
        Option(
            "--ps",
            "--processes",
            help=(
                "Num. processes for multprocess generation."
                " One process per file is used for multi-file generation"
            ),
        ),
    ] = None,
    sanitize: Annotated[
        str,
        Option(
            "--sanitize",
            help="RDKit sanitization operations to perform ('all' or 'minimal')",
        ),
    ] = "all",
    skip_invalid: Annotated[
        bool,
        Option(
            "--skip-invalid/--no-skip-invalid",
            help=(
                "Skip invalid smiles."
                " If False, an error is raised on invalid smiles. If True they are"
                " silently skipped (this is be more memory intensive, especially for"
                " parallel processing)"
            ),
        ),
    ] = False,
    tab_separated: Annotated[
        bool,
        Option(
            "--tab-sep/--no-tab-sep",
            help="Whether the smiles file has the format <smiles><tab><field><tab>...",
        ),
    ] = False,
    replace_dummy_atoms: Annotated[
        bool,
        Option(
            "--replace-dummy/--no-replace-dummy",
            help="Whether to replace dummy atoms such as [U], [Np], etc. used in synthon spaces",  # noqa
            hidden=True,
        ),
    ] = False,
) -> None:
    r"""Generate a `*.npy` fingerprints file from one or more `*.smi` smiles files

    By default this function runs in parallel and uses all available CPUs. In order to
    use the memory efficient BitBIRCH u8 algorithm you should keep the defaults:
    --dtype=uint8 and --pack
    """
    import numpy as np

    from bblean._console import get_console
    from bblean.utils import _num_avail_cpus
    from bblean.fingerprints import _FingerprintFileCreator, _FingerprintArrayFiller
    from bblean.smiles import (
        calc_num_smiles,
        _iter_ranges_and_smiles_batches,
        _iter_idxs_and_smiles_batches,
    )

    # Force forkserver since rdkit may use threads, and fork is unsafe with threads
    mp_context = mp.get_context("forkserver" if sys.platform == "linux" else None)

    console = get_console(silent=not verbose)

    if smiles_paths is None:
        smiles_paths = list(Path.cwd().glob("*.smi"))
    if not smiles_paths:
        console.print("No *.smi files found", style="red")
        raise Abort()

    smiles_num = calc_num_smiles(smiles_paths)

    def parse_num_per_batch(
        smiles_num: int, parts: int | None, max_fps_per_file: int | None
    ) -> tuple[int, int, int | None]:
        digits: int | None
        if parts is not None and max_fps_per_file is None:
            num_per_batch = math.ceil(smiles_num / parts)
            digits = len(str(parts))
        elif parts is None and max_fps_per_file is not None:
            num_per_batch = max_fps_per_file
            parts = math.ceil(smiles_num / max_fps_per_file)
            digits = len(str(parts))
        elif parts is None and max_fps_per_file is None:
            parts = 1
            num_per_batch = math.ceil(smiles_num / parts)
            digits = None
        else:
            raise ValueError("parts and max_fps_per_file are mutually exclusive")
        return parts, num_per_batch, digits

    try:
        parts, num_per_batch, digits = parse_num_per_batch(
            smiles_num, parts, max_fps_per_file
        )
    except ValueError:
        console.print(
            "'--max-fps-per-file' and '--num-parts' are mutually exclusive",
            style="red",
        )
        raise Abort() from None
    if out_dir is None:
        out_dir = Path.cwd()
    out_dir.mkdir(exist_ok=True)
    out_dir = out_dir.resolve()

    # Pass 2: build the molecules
    unique_id = format(random.getrandbits(32), "08x")
    if out_name is None:
        # Save the fingerprints as a NumPy array
        out_name = f"{'packed-' if pack else ''}fps-{dtype}-{kind}-{unique_id}"
    else:
        # Strip suffix
        if out_name.endswith(".npy"):
            out_name = out_name[:-4]

    if num_ps is None:
        # Get the number of cores *available for use for this process*
        # bound by the number of parts to avoid spawning useless processes
        if parts == 1:
            num_ps = _num_avail_cpus()
        else:
            num_ps = min(_num_avail_cpus(), parts)
    create_fp_file = _FingerprintFileCreator(
        dtype,
        out_dir,
        out_name,
        digits,
        pack,
        kind,
        fp_size,
        sanitize=sanitize,
        skip_invalid=skip_invalid,
        verbose=verbose,
    )
    timer = Timer()
    timer.init_timing("total")
    if parts > 1 and num_ps is not None and num_ps > 1:
        # Multiprocessing version, 1 process per file
        with console.status(
            f"[italic]Generating fingerprints ({parts} files, parallel, {num_ps} procs.) ...[/italic]",  # noqa:E501
            spinner="dots",
        ):
            with mp_context.Pool(processes=num_ps) as pool:
                pool.map(
                    create_fp_file,
                    _iter_idxs_and_smiles_batches(
                        smiles_paths, num_per_batch, tab_separated, replace_dummy_atoms
                    ),
                )
        timer.end_timing("total", console, indent=False)
        stem = out_name.split(".")[0]
        console.print(f"Finished. Outputs written to {str(out_dir / stem)}.<idx>.npy")
        return

    # Parallel or serial, single file version
    msg = "parallel" if num_ps > 1 else "serial"
    with console.status(
        f"[italic]Generating fingerprints ({parts} files, {msg}, {num_ps} procs.) ...[/italic]",  # noqa:E501
        spinner="dots",
    ):
        if pack:
            out_dim = (fp_size + 7) // 8
        else:
            out_dim = fp_size
        shmem_size = smiles_num * out_dim * np.dtype(dtype).itemsize
        fps_shmem = shmem.SharedMemory(create=True, size=shmem_size)
        invalid_mask_shmem = shmem.SharedMemory(create=True, size=smiles_num)
        fps_array_filler = _FingerprintArrayFiller(
            shmem_name=fps_shmem.name,
            invalid_mask_shmem_name=invalid_mask_shmem.name,
            kind=kind,
            fp_size=fp_size,
            num_smiles=smiles_num,
            dtype=dtype,
            pack=pack,
            sanitize=sanitize,
            skip_invalid=skip_invalid,
        )
        if num_ps > 1 and parts == 1:
            # Split into batches anyways if we have a single batch but multiple
            # processes
            _, num_per_batch, _ = parse_num_per_batch(
                smiles_num, num_ps, max_fps_per_file
            )
        with mp_context.Pool(processes=num_ps) as pool:
            pool.starmap(
                fps_array_filler,
                _iter_ranges_and_smiles_batches(
                    smiles_paths, num_per_batch, tab_separated, replace_dummy_atoms
                ),
            )
        fps = np.ndarray((smiles_num, out_dim), dtype=dtype, buffer=fps_shmem.buf)
        mask = np.ndarray((smiles_num,), dtype=np.bool, buffer=invalid_mask_shmem.buf)
        if skip_invalid:
            prev_num = len(fps)
            fps = np.delete(fps, mask, axis=0)
            new_num = len(fps)
            console.print(f"Generated {new_num} fingerprints")
            console.print(f"Skipped {prev_num - new_num} invalid smiles")
            invalid_name = f"invalid-{unique_id}.npy"
            console.print(
                f"Invalid smiles idxs written to {str(out_dir / invalid_name)}"
            )
            np.save(out_dir / f"invalid-{unique_id}.npy", mask.nonzero()[0].reshape(-1))

        np.save(
            out_dir / out_name,
            fps,
        )
        del mask
        del fps
        # Cleanup
        fps_shmem.unlink()
        invalid_mask_shmem.unlink()
    timer.end_timing("total", console, indent=False)
    console.print(f"Finished. Outputs written to {str(out_dir / out_name)}.npy")


@app.command("fps-split", rich_help_panel="Fingerprints")
def _split_fps(
    input_: Annotated[
        Path,
        Argument(help="`*.npy` file with fingerprints"),
    ],
    out_dir: Annotated[
        Path | None,
        Option("-o", "--out-dir", show_default=False),
    ] = None,
    parts: Annotated[
        int | None,
        Option(
            "-n",
            "--num-parts",
            help="Num. of parts to split file into. Mutually exclusive with --max-fps",
            show_default=False,
        ),
    ] = None,
    max_fps_per_file: Annotated[
        int | None,
        Option(
            "-m",
            "--max-fps",
            help="Max. number of fps per file. Mutually exclusive with --num-parts",
            show_default=False,
        ),
    ] = None,
) -> None:
    r"""Split a `*.npy` fingerprint file into multiple `*.npy` files

    Usage to split into multiple files with a max number of fps each (e.g. 10k) is `bb
    split-fps --max-fps 10_000 ./fps.npy --out-dir ./split`. To split into a pre-defined
    number of parts (e.g. 10) `bb split-fps --num-parts 10 ./fps.npy --out-dir ./split`.
    """
    from bblean._console import get_console
    import numpy as np

    console = get_console()
    if parts is not None and parts < 2:
        console.print("Num must be >= 2", style="red")
        raise Abort()
    fps = np.load(input_, mmap_mode="r")
    if parts is not None and max_fps_per_file is None:
        num_per_batch = math.ceil(fps.shape[0] / parts)
        digits = len(str(parts))
    elif parts is None and max_fps_per_file is not None:
        num_per_batch = max_fps_per_file
        digits = len(str(math.ceil(fps.shape[0] / max_fps_per_file)))
    else:
        console.print(
            "One and only one of '--max-fps' and '--num-parts' required", style="red"
        )
        raise Abort()

    stem = input_.name.split(".")[0]
    with console.status("[italic]Splitting fingerprints...[/italic]", spinner="dots"):
        i = -1
        for i, batch in enumerate(batched(fps, num_per_batch)):
            suffixes = input_.suffixes
            name = f"{stem}{''.join(suffixes[:-1])}.{str(i).zfill(digits)}.npy"

            # Generate out dir when first fp file is being saved
            if out_dir is None:
                out_dir = Path.cwd() / stem
            out_dir.mkdir(exist_ok=True)
            out_dir = out_dir.resolve()

            np.save(out_dir / name, batch)

        if i == -1:
            console.print("Warning: No fingerprints written", style="yellow")
            return
    console.print(
        f"Finished. Outputs written to {str(tp.cast(Path, out_dir) / stem)}.<idx>.npy"
    )


@app.command("fps-shuffle", rich_help_panel="Fingerprints")
def _shuffle_fps(
    in_path: Annotated[
        Path,
        Argument(help="`*.npy` file with fingerprints, or dir with `*.npy` files"),
    ],
    out_dir: Annotated[
        Path | None,
        Option("-o", "--out-dir", show_default=False),
    ] = None,
    seed: Annotated[
        int | None,
        Option("--seed", hidden=True, rich_help_panel="Debug"),
    ] = None,
    save_shuffle_idxs: Annotated[
        bool,
        Option("--save-shuffle-idxs/--no-save-shuffle-idxs"),
    ] = True,
) -> None:
    """Shuffle a fingerprints file

    This function is not optimized and as such may have high RAM usage. It is
    meant for testing purposes only"""
    import numpy as np
    from bblean._console import get_console

    console = get_console()

    console = get_console()
    if in_path.is_dir():
        files = sorted(
            f for f in in_path.glob("*.npy") if not f.stem.endswith(".indices")
        )
    else:
        files = [in_path]
    if out_dir is None:
        out_dir = Path.cwd()
    out_dir.mkdir(exist_ok=True)
    out_dir = out_dir.resolve()
    for f in files:
        with console.status(
            "[italic]Shuffling fingerprints...[/italic]", spinner="dots"
        ):
            fps = np.load(f)
            stem = f.stem
            rng = np.random.default_rng(seed)
            shuffle_idxs = rng.permutation(fps.shape[0])
            fps = fps[shuffle_idxs]
            stem = f"shuffled-{stem}"
            np.save(out_dir / f"{stem}.npy", fps)
            if save_shuffle_idxs:
                np.save(out_dir / f"{stem}.indices.npy", shuffle_idxs)
        if save_shuffle_idxs:
            console.print(
                f"Finished. Outputs written to {str(out_dir / stem)}.npy and {str(out_dir / stem)}.indices.npy"  # noqa
            )
        else:
            console.print(f"Finished. Outputs written to {str(out_dir / stem)}.npy")


@app.command("fps-merge", rich_help_panel="Fingerprints")
def _merge_fps(
    in_dir: Annotated[
        Path,
        Argument(help="Directory with input `*.npy` files with packed fingerprints"),
    ],
    out_dir: Annotated[
        Path | None,
        Option("-o", "--out-dir", show_default=False),
    ] = None,
) -> None:
    r"""Merge a dir with multiple `*.npy` fingerprint file into a single `*.npy` file"""
    from bblean._console import get_console
    import numpy as np

    console = get_console()

    if out_dir is None:
        out_dir = Path.cwd()
    out_dir.mkdir(exist_ok=True)
    out_dir = out_dir.resolve()
    arrays = []
    with console.status("[italic]Merging fingerprints...[/italic]", spinner="dots"):
        stem = None
        for f in sorted(in_dir.glob("*.npy")):
            if f.stem.endswith(".indices"):
                continue
            if stem is None:
                stem = f.name.split(".")[0]
            elif stem != f.name.split(".")[0]:
                raise ValueError(
                    "Name convention must be <name>.<idx>.npy"
                    " with all files having the same <name>"
                )
            arrays.append(np.load(f))
        if stem is None:
            console.print("No *.npy files found")
            return
        np.save(out_dir / stem, np.concatenate(arrays))
    console.print(f"Finished. Outputs written to {str(out_dir / stem)}.npy")


@app.command("fps-sort", rich_help_panel="Fingerprints")
def _sort_fps(
    in_path: Annotated[
        Path,
        Argument(help="`*.npy` file with fingerprints, or dir with `*.npy` files"),
    ],
    out_dir: Annotated[
        Path | None,
        Option("-o", "--out-dir", show_default=False),
    ] = None,
    seed: Annotated[
        int | None,
        Option("--seed", hidden=True, rich_help_panel="Debug"),
    ] = None,
    input_is_packed: Annotated[
        bool,
        Option("--packed-input/--unpacked-input", rich_help_panel="Advanced"),
    ] = True,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
    save_sort_idxs: Annotated[
        bool,
        Option("--save-sort-idxs/--no-save-sort-idxs"),
    ] = True,
) -> None:
    r"""Sort a fingerprints file by popcount"""
    import numpy as np
    from bblean._py_similarity import _popcount
    from bblean._console import get_console
    from bblean.fingerprints import pack_fingerprints

    # Note that n_features is not used here even if input_is_packed is True,
    # it is added for API homogeneity

    console = get_console()
    if in_path.is_dir():
        files = sorted(
            f for f in in_path.glob("*.npy") if not f.stem.endswith(".indices")
        )
    else:
        files = [in_path]
    if out_dir is None:
        out_dir = Path.cwd()
    out_dir.mkdir(exist_ok=True)
    out_dir = out_dir.resolve()
    for f in files:
        with console.status(
            "[italic]Sorting fingerprints by popcount...[/italic]", spinner="dots"
        ):
            fps = np.load(f)
            stem = f.stem
            if not input_is_packed:
                packed_fps = pack_fingerprints(fps)
            else:
                packed_fps = fps
            counts = _popcount(packed_fps)
            sort_idxs = np.argsort(counts)
            fps = fps[sort_idxs]
            stem = f"sorted-{stem}"
            np.save(out_dir / f"{stem}.npy", fps)
            if save_sort_idxs:
                np.save(out_dir / f"{stem}.indices.npy", sort_idxs)

        if save_sort_idxs:
            console.print(
                f"Finished. Outputs written to {str(out_dir / stem)}.npy and {str(out_dir / stem)}.indices.npy"  # noqa
            )
        else:
            console.print(f"Finished. Outputs written to {str(out_dir / stem)}.npy")


@app.command("fps-unpack", rich_help_panel="Fingerprints")
def _unpack_fps(
    in_path: Annotated[
        Path,
        Argument(help="`*.npy` file with fingerprints, or dir with `*.npy` files"),
    ],
    out_dir: Annotated[
        Path | None,
        Option("-o", "--out-dir", show_default=False),
    ] = None,
    n_features: Annotated[
        int | None,
        Option(
            "--n-features",
            help="Number of features in the fingerprints."
            " Only for packed inputs *if it is not a multiple of 8*."
            " Not required for typical fingerprint sizes (e.g. 2048, 1024)",
            rich_help_panel="Advanced",
        ),
    ] = None,
) -> None:
    r"""Unpack a fingerprints file"""
    import numpy as np
    from bblean.fingerprints import unpack_fingerprints
    from bblean._console import get_console

    console = get_console()

    if in_path.is_dir():
        files = sorted(
            f for f in in_path.glob("*.npy") if not f.stem.endswith(".indices")
        )
    else:
        files = [in_path]
    if out_dir is None:
        out_dir = Path.cwd()
    out_dir.mkdir(exist_ok=True)
    out_dir = out_dir.resolve()
    for f in files:
        with console.status(
            "[italic]Unpacking fingerprints...[/italic]", spinner="dots"
        ):
            fps = np.load(f)
            stem = f.stem
            if "unpacked" in stem:
                warnings.warn(
                    "The fingerprints file name containes 'unpacked',"
                    " make sure the file contains packed fps"
                )
                stem = f"unpacked-{stem}"
            elif "packed" in stem:
                stem = stem.replace("packed", "unpacked")
            else:
                stem = f"unpacked-{stem}"
            unpacked_fps = unpack_fingerprints(fps, n_features)
            np.save(out_dir / f"{stem}.npy", unpacked_fps)
        console.print(f"Finished. Outputs written to {str(out_dir / stem)}.npy")


@app.command("fps-pack", rich_help_panel="Fingerprints")
def _pack_fps(
    in_path: Annotated[
        Path,
        Argument(help="`*.npy` file with fingerprints, or dir with `*.npy` files"),
    ],
    out_dir: Annotated[
        Path | None,
        Option("-o", "--out-dir", show_default=False),
    ] = None,
) -> None:
    r"""Pack a fingerprints file"""
    import numpy as np
    from bblean.fingerprints import pack_fingerprints
    from bblean._console import get_console

    console = get_console()

    if in_path.is_dir():
        files = sorted(
            f for f in in_path.glob("*.npy") if not f.stem.endswith(".indices")
        )
    else:
        files = [in_path]
    if out_dir is None:
        out_dir = Path.cwd()
    out_dir.mkdir(exist_ok=True)
    out_dir = out_dir.resolve()
    for f in files:
        with console.status("[italic]Packing fingerprints...[/italic]", spinner="dots"):
            fps = np.load(f)
            stem = f.stem
            if "packed" in stem and "unpacked" not in "stem":
                msg = (
                    "The fingerprints file name containes 'packed',"
                    " make sure the file contains packed fps"
                )
                warnings.warn(msg)
                stem = f"packed-{stem}"
            elif "unpacked" in stem:
                stem = stem.replace("unpacked", "packed")
            else:
                stem = f"packed-{stem}"
            unpacked_fps = pack_fingerprints(fps)
            np.save(out_dir / f"{stem}.npy", unpacked_fps)
        console.print(f"Finished. Outputs written to {str(out_dir / stem)}.npy")



# BitBIRCH-Lean Python Package: An open-source clustering module based on iSIM.
#
# If you find this software useful please cite the following articles:
# - BitBIRCH: efficient clustering of large molecular libraries:
#   https://doi.org/10.1039/D5DD00030K
# - BitBIRCH Clustering Refinement Strategies:
#   https://doi.org/10.1021/acs.jcim.5c00627
# - BitBIRCh-Lean:
#   (preprint) https://www.biorxiv.org/content/10.1101/2025.10.22.684015v1
#
# Copyright (C) 2025  The Miranda-Quintana Lab and other BitBirch developers, comprised
# exclusively by:
# - Ramon Alain Miranda Quintana <ramirandaq@gmail.com>, <quintana@chem.ufl.edu>
# - Krisztina Zsigmond <kzsigmond@ufl.edu>
# - Ignacio Pickering <ipickering@chem.ufl.edu>
# - Kenneth Lopez Perez <klopezperez@chem.ufl.edu>
# - Miroslav Lzicar <miroslav.lzicar@deepmedchem.com>
#
# Authors of this file are:
# - Ramon Alain Miranda Quintana <ramirandaq@gmail.com>, <quintana@chem.ufl.edu>
# - Ignacio Pickering <ipickering@chem.ufl.edu>
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# version 3 (SPDX-License-Identifier: GPL-3.0-only).
#
# Portions of ./bblean/bitbirch.py are licensed under the BSD 3-Clause License
# Copyright (c) 2007-2024 The scikit-learn developers. All rights reserved.
# (SPDX-License-Identifier: BSD-3-Clause). Copies or reproductions of code in the
# ./bblean/bitbirch.py file must in addition adhere to the BSD-3-Clause license terms. A
# copy of the BSD-3-Clause license can be located at the root of this repository, under
# ./LICENSES/BSD-3-Clause.txt.
#
# Portions of ./bblean/bitbirch.py were previously licensed under the LGPL 3.0
# license (SPDX-License-Identifier: LGPL-3.0-only), they are relicensed in this program
# as GPL-3.0, with permission of all original copyright holders:
# - Ramon Alain Miranda Quintana <ramirandaq@gmail.com>, <quintana@chem.ufl.edu>
# - Vicky (Vic) Jung <jungvicky@ufl.edu>
# - Kenneth Lopez Perez <klopezperez@chem.ufl.edu>
# - Kate Huddleston <kdavis2@chem.ufl.edu>
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program. This copy can be located at the root of this repository, under
# ./LICENSES/GPL-3.0-only.txt.  If not, see <http://www.gnu.org/licenses/gpl-3.0.html>.
r"""Multi-round BitBirch workflow for clustering huge datasets in parallel"""
import sys
import math
import pickle
import typing as tp
import multiprocessing as mp
from pathlib import Path

from rich.console import Console
import numpy as np
from numpy.typing import NDArray


from bblean._console import get_console
from bblean._timer import Timer
from bblean._config import DEFAULTS
from bblean.utils import batched
from bblean.bitbirch import BitBirch
from bblean.fingerprints import _get_fps_file_num

__all__ = ["run_multiround_bitbirch"]


# Save a list of numpy arrays into a single array in a streaming fashion, avoiding
# stacking them in memory
def _numpy_streaming_save(
    fp_list: list[NDArray[np.integer]] | NDArray[np.integer], path: Path | str
) -> None:
    first_arr = np.ascontiguousarray(fp_list[0])
    header = np.lib.format.header_data_from_array_1_0(first_arr)
    header["shape"] = (len(fp_list), len(first_arr))
    path = Path(path)
    if not path.suffix:
        path = path.with_suffix(".npy")
    with open(path, "wb") as f:
        np.lib.format.write_array_header_1_0(f, header)
        for arr in fp_list:
            np.ascontiguousarray(arr).tofile(f)


# Glob and sort by uint bits and label, if a console is passed then the number of output
# files is printed
def _get_prev_round_buf_and_mol_idxs_files(
    path: Path, round_idx: int, console: Console | None = None
) -> list[tuple[Path, Path]]:
    path = Path(path)
    # TODO: Important: What should be the logic for batching? currently there doesn't
    # seem to be much logic for grouping the files
    buf_files = sorted(path.glob(f"round-{round_idx - 1}-bufs*.npy"))
    idx_files = sorted(path.glob(f"round-{round_idx - 1}-idxs*.pkl"))
    if console is not None:
        console.print(f"    - Collected {len(buf_files)} buffer-index file pairs")
    return list(zip(buf_files, idx_files))


def _sort_batch(b: tp.Sequence[tuple[Path, Path]]) -> tuple[tuple[Path, Path], ...]:
    return tuple(
        sorted(
            b,
            key=lambda b: int(b[0].name.split("uint")[-1].split(".")[0]),
            reverse=True,
        )
    )


def _chunk_file_pairs_in_batches(
    file_pairs: tp.Sequence[tuple[Path, Path]],
    bin_size: int,
    console: Console | None = None,
) -> list[tuple[str, tuple[tuple[Path, Path], ...]]]:
    z = len(str(math.ceil(len(file_pairs) / bin_size)))
    # Within each batch, sort the files by starting with the uint16 files, followed by
    # uint8 files, this helps that (approximately) the largest clusters are fitted first
    # which may improve final cluster quality
    batches = [
        (str(i).zfill(z), _sort_batch(b))
        for i, b in enumerate(batched(file_pairs, bin_size))
    ]
    if console is not None:
        console.print(f"    - Chunked files into {len(batches)} batches")
    return batches


def _save_bufs_and_mol_idxs(
    out_dir: Path,
    fps_bfs: dict[str, tp.Any],
    mols_bfs: dict[str, tp.Any],
    label: str,
    round_idx: int,
) -> None:
    for dtype, buf_list in fps_bfs.items():
        suffix = f".label-{label}-{dtype.replace('8', '08')}"
        _numpy_streaming_save(buf_list, out_dir / f"round-{round_idx}-bufs{suffix}.npy")
        with open(out_dir / f"round-{round_idx}-idxs{suffix}.pkl", mode="wb") as f:
            pickle.dump(mols_bfs[dtype], f)


class _InitialRound:
    def __init__(
        self,
        branching_factor: int,
        threshold: float,
        tolerance: float,
        out_dir: Path | str,
        refinement_before_midsection: str,
        refine_threshold_change: float,
        refine_merge_criterion: str,
        n_features: int | None = None,
        max_fps: int | None = None,
        merge_criterion: str = DEFAULTS.merge_criterion,
        input_is_packed: bool = True,
    ) -> None:
        self.n_features = n_features
        self.refinement_before_midsection = refinement_before_midsection
        if refinement_before_midsection not in ["full", "split", "none"]:
            raise ValueError(f"Unknown refinement kind {refinement_before_midsection}")
        self.branching_factor = branching_factor
        self.threshold = threshold
        self.tolerance = tolerance
        self.out_dir = Path(out_dir)
        self.max_fps = max_fps
        self.merge_criterion = merge_criterion
        self.refine_merge_criterion = refine_merge_criterion
        self.input_is_packed = input_is_packed
        self.refine_threshold_change = refine_threshold_change

    def __call__(self, file_info: tuple[str, Path, int, int]) -> None:
        file_label, fp_file, start_idx, end_idx = file_info

        # First fit the fps in each process, in parallel.
        # `reinsert_indices` required to keep track of mol idxs in different processes.
        tree = BitBirch(
            branching_factor=self.branching_factor,
            threshold=self.threshold,
            merge_criterion=self.merge_criterion,
        )

        range_ = range(start_idx, end_idx)
        tree.fit(
            fp_file,
            reinsert_indices=range_,
            n_features=self.n_features,
            input_is_packed=self.input_is_packed,
            max_fps=self.max_fps,
        )
        # Extract the BitFeatures of the leaves, breaking the largest cluster(s) apart,
        # to prepare for refinement
        tree.delete_internal_nodes()
        if self.refinement_before_midsection == "none":
            fps_bfs, mols_bfs = tree._bf_to_np()
        elif self.refinement_before_midsection in ["split", "full"]:
            fps_bfs, mols_bfs = tree._bf_to_np_refine(fp_file, initial_mol=start_idx)
            if self.refinement_before_midsection == "full":
                # Finish the first refinement step internally in this round
                tree.reset()
                tree.set_merge(
                    merge_criterion=self.refine_merge_criterion,
                    tolerance=self.tolerance,
                    threshold=self.threshold + self.refine_threshold_change,
                )
                for bufs, mol_idxs in zip(fps_bfs.values(), mols_bfs.values()):
                    tree._fit_buffers(bufs, reinsert_index_seqs=mol_idxs)
                    del mol_idxs
                    del bufs

                tree.delete_internal_nodes()
                fps_bfs, mols_bfs = tree._bf_to_np()

        _save_bufs_and_mol_idxs(self.out_dir, fps_bfs, mols_bfs, file_label, 1)


class _TreeMergingRound:
    def __init__(
        self,
        branching_factor: int,
        threshold: float,
        tolerance: float,
        round_idx: int,
        out_dir: Path | str,
        split_largest_cluster: bool,
        merge_criterion: str,
        all_fp_paths: tp.Sequence[Path] = (),
    ) -> None:
        self.all_fp_paths = list(all_fp_paths)
        self.branching_factor = branching_factor
        self.threshold = threshold
        self.tolerance = tolerance
        self.round_idx = round_idx
        self.out_dir = Path(out_dir)
        self.split_largest_cluster = split_largest_cluster
        self.merge_criterion = merge_criterion

    def __call__(self, batch_info: tuple[str, tp.Sequence[tuple[Path, Path]]]) -> None:
        batch_label, batch_path_pairs = batch_info
        tree = BitBirch(
            branching_factor=self.branching_factor,
            threshold=self.threshold,
            merge_criterion=self.merge_criterion,
            tolerance=self.tolerance,
        )
        # Rebuild a tree, inserting all BitFeatures from the corresponding batch
        for buf_path, idx_path in batch_path_pairs:
            with open(idx_path, "rb") as f:
                mol_idxs = pickle.load(f)
            tree._fit_buffers(buf_path, reinsert_index_seqs=mol_idxs)
            del mol_idxs

        # Either do a refinement step, or fetch and save the bufs and idxs for the next
        # round
        tree.delete_internal_nodes()
        if self.split_largest_cluster:
            fps_bfs, mols_bfs = tree._bf_to_np_refine(self.all_fp_paths)
        else:
            fps_bfs, mols_bfs = tree._bf_to_np()
        _save_bufs_and_mol_idxs(
            self.out_dir, fps_bfs, mols_bfs, batch_label, self.round_idx
        )


class _FinalTreeMergingRound(_TreeMergingRound):
    def __init__(
        self,
        branching_factor: int,
        threshold: float,
        tolerance: float,
        merge_criterion: str,
        out_dir: Path | str,
        save_tree: bool,
        save_centroids: bool,
    ) -> None:
        super().__init__(
            branching_factor,
            threshold,
            tolerance,
            -1,
            out_dir,
            False,
            merge_criterion,
            (),
        )
        self.save_tree = save_tree
        self.save_centroids = save_centroids

    def __call__(self, batch_info: tuple[str, tp.Sequence[tuple[Path, Path]]]) -> None:
        batch_path_pairs = batch_info[1]
        tree = BitBirch(
            branching_factor=self.branching_factor,
            threshold=self.threshold,
            merge_criterion=self.merge_criterion,
            tolerance=self.tolerance,
        )
        # Rebuild a tree, inserting all BitFeatures from the corresponding batch
        for buf_path, idx_path in batch_path_pairs:
            with open(idx_path, "rb") as f:
                mol_idxs = pickle.load(f)
            tree._fit_buffers(buf_path, reinsert_index_seqs=mol_idxs)
            del mol_idxs

        # Save clusters and exit
        if self.save_tree:
            # TODO: Find alternative solution
            tree.save(self.out_dir / "bitbirch.pkl")
        tree.delete_internal_nodes()
        if self.save_centroids:
            output = tree.get_centroids_mol_ids()
            with open(self.out_dir / "clusters.pkl", mode="wb") as f:
                pickle.dump(output["mol_ids"], f)
            with open(self.out_dir / "cluster-centroids-packed.pkl", mode="wb") as f:
                pickle.dump(output["centroids"], f)
        else:
            with open(self.out_dir / "clusters.pkl", mode="wb") as f:
                pickle.dump(tree.get_cluster_mol_ids(), f)


# Create a list of tuples of labels, file paths and start-end idxs
def _get_files_range_tuples(
    files: tp.Sequence[Path],
) -> list[tuple[str, Path, int, int]]:
    running_idx = 0
    files_info = []
    z = len(str(len(files)))
    for i, file in enumerate(files):
        start_idx = running_idx
        end_idx = running_idx + _get_fps_file_num(file)
        files_info.append((str(i).zfill(z), file, start_idx, end_idx))
        running_idx = end_idx
    return files_info


# NOTE: 'full_refinement_before_midsection' indicates if the refinement of the batches
# is fully done after the tree-merging rounds, or if the data is only split before the
# tree-merging rounds
def run_multiround_bitbirch(
    input_files: tp.Sequence[Path],
    out_dir: Path,
    n_features: int | None = None,
    input_is_packed: bool = True,
    num_initial_processes: int = 10,
    num_midsection_processes: int | None = None,
    initial_merge_criterion: str = DEFAULTS.merge_criterion,
    branching_factor: int = DEFAULTS.branching_factor,
    threshold: float = DEFAULTS.threshold,
    midsection_threshold_change: float = DEFAULTS.refine_threshold_change,
    tolerance: float = DEFAULTS.tolerance,
    # Advanced
    num_midsection_rounds: int = 1,
    bin_size: int = 10,
    max_tasks_per_process: int = 1,
    refinement_before_midsection: str = "full",
    split_largest_after_each_midsection_round: bool = False,
    midsection_merge_criterion: str = DEFAULTS.refine_merge_criterion,
    final_merge_criterion: str | None = None,
    mp_context: tp.Any = None,
    save_tree: bool = False,
    save_centroids: bool = True,
    # Debug
    max_fps: int | None = None,
    verbose: bool = False,
    cleanup: bool = True,
) -> Timer:
    r"""Perform (possibly parallel) multi-round BitBirch clustering

    .. warning::

        The functionality provided by this function is stable, but its API
        (the arguments it takes and its return values) may change in the future.
    """
    if final_merge_criterion is None:
        final_merge_criterion = midsection_merge_criterion
    if mp_context is None:
        mp_context = mp.get_context("forkserver" if sys.platform == "linux" else None)
    # Returns timing and for the different rounds
    # TODO: Also return peak-rss
    console = get_console(silent=not verbose)

    if num_midsection_processes is None:
        num_midsection_processes = num_initial_processes
    else:
        # Sanity check
        if num_midsection_processes > num_initial_processes:
            raise ValueError("Num. midsection procs. must be <= num. initial processes")

    # Common params to all rounds BitBIRCH
    common_kwargs: dict[str, tp.Any] = dict(
        branching_factor=branching_factor,
        tolerance=tolerance,
        out_dir=out_dir,
    )
    timer = Timer()
    timer.init_timing("total")

    # Get starting and ending idxs for each file, and collect them into tuples
    files_range_tuples = _get_files_range_tuples(input_files)  # correct
    num_files = len(input_files)

    # Initial round of clustering
    round_idx = 1
    timer.init_timing(f"round-{round_idx}")
    console.print(f"(Initial) Round {round_idx}: Cluster initial batch of fingerprints")

    initial_fn = _InitialRound(
        n_features=n_features,
        refinement_before_midsection=refinement_before_midsection,
        max_fps=max_fps,
        merge_criterion=initial_merge_criterion,
        input_is_packed=input_is_packed,
        threshold=threshold,
        refine_merge_criterion=midsection_merge_criterion,
        refine_threshold_change=midsection_threshold_change,
        **common_kwargs,
    )
    num_ps = min(num_initial_processes, num_files)
    console.print(f"    - Processing {num_files} inputs with {num_ps} processes")
    with console.status("[italic]BitBirching...[/italic]", spinner="dots"):
        if num_ps == 1:
            for tup in files_range_tuples:
                initial_fn(tup)
        else:
            with mp_context.Pool(
                processes=num_ps, maxtasksperchild=max_tasks_per_process
            ) as pool:
                pool.map(initial_fn, files_range_tuples)

    timer.end_timing(f"round-{round_idx}", console)
    console.print_peak_mem(out_dir)

    # Mid-section "Tree-Merging" rounds of clustering
    for _ in range(num_midsection_rounds):
        round_idx += 1
        timer.init_timing(f"round-{round_idx}")
        console.print(f"(Midsection) Round {round_idx}: Re-clustering in chunks")

        file_pairs = _get_prev_round_buf_and_mol_idxs_files(out_dir, round_idx, console)
        batches = _chunk_file_pairs_in_batches(file_pairs, bin_size, console)
        merging_fn = _TreeMergingRound(
            round_idx=round_idx,
            all_fp_paths=input_files,
            split_largest_cluster=split_largest_after_each_midsection_round,
            merge_criterion=midsection_merge_criterion,
            threshold=threshold + midsection_threshold_change,
            **common_kwargs,
        )
        num_ps = min(num_midsection_processes, len(batches))
        console.print(f"    - Processing {len(batches)} inputs with {num_ps} processes")
        with console.status("[italic]BitBirching...[/italic]", spinner="dots"):
            if num_ps == 1:
                for batch_info in batches:
                    merging_fn(batch_info)
            else:
                with mp_context.Pool(
                    processes=num_ps, maxtasksperchild=max_tasks_per_process
                ) as pool:
                    pool.map(merging_fn, batches)

        timer.end_timing(f"round-{round_idx}", console)
        console.print_peak_mem(out_dir)

    # Final "Tree-Merging" round of clustering
    round_idx += 1
    timer.init_timing(f"round-{round_idx}")
    console.print(f"(Final) Round {round_idx}: Final round of clustering")
    file_pairs = _get_prev_round_buf_and_mol_idxs_files(out_dir, round_idx, console)

    final_fn = _FinalTreeMergingRound(
        save_tree=save_tree,
        save_centroids=save_centroids,
        merge_criterion=final_merge_criterion,
        threshold=threshold + midsection_threshold_change,
        **common_kwargs,
    )
    with console.status("[italic]BitBirching...[/italic]", spinner="dots"):
        final_fn(("", file_pairs))

    timer.end_timing(f"round-{round_idx}", console)
    console.print_peak_mem(out_dir)
    # Remove intermediate files
    if cleanup:
        for f in out_dir.glob("round-*.npy"):
            f.unlink()
        for f in out_dir.glob("round-*.pkl"):
            f.unlink()
    console.print()
    timer.end_timing("total", console, indent=False)
    return timer



# type: ignore
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import plotly.graph_objects as go
from collections import defaultdict
from rdkit.Chem import rdFMCS, SaltRemover
from rdkit import Chem
from rdkit.Chem import Draw
import bblean.similarity as iSIM

# This file includes the functions for the visualizations of the best practices paper.


def clusters_pop_plot(
    clusters: list,
    save_path: str = None,
):
    """Plot the population of each cluster as a bar chart.

    Args:
        clusters (list): List of cluster sizes.
        save_path (str, optional): Path to save the plot. Defaults to None.
    """

    # Calculate the counts of the populations
    lenghts = [len(c) for c in clusters]
    n_1000 = sum(1 for lenght in lenghts if lenght > 1000)
    n_100 = sum(1 for lenght in lenghts if lenght > 100)
    n_10 = sum(1 for lenght in lenghts if lenght > 10)
    n_1 = sum(1 for lenght in lenghts if lenght > 1)
    n_0 = sum(1 for lenght in lenghts if lenght > 0)

    plt.figure(figsize=(3, 4))
    plt.bar("Num_cluster", n_0, label=">0", color="blue")
    plt.bar("Num_cluster", n_1, label=">1", color="orange")
    plt.bar("Num_cluster", n_10, label=">10", color="gray")
    plt.bar("Num_cluster", n_100, label=">100", color="green")
    plt.bar("Num_cluster", n_1000, label=">1000", color="red")
    plt.legend()
    plt.ylabel("Number of Clusters")
    plt.tight_layout()

    if save_path:
        plt.savefig(save_path, dpi=400)
    else:
        plt.show()


def sankey_cluster_flow(
    clusters_before,
    clusters_after,
    min_flow=1,
    top_n=20,
    max_flow=None,
    title=None,
    show_counts=True,
):
    """Plot a Sankey diagram showing flow from clusters_before to clusters_after.
    Parameters:
    ---------------------
    clusters_before / clusters_after:
    list-of-lists of integer molecule indices.
    min_flow:
    minimum number of molecules for a link to be considered (per before->after link).
    top_n:
    show flows into the top_n largest clusters in clusters_after (by size).
    max_flow:
    if set, only show links with count <= max_flow (useful to inspect small flows).
    title:
    optional title for the figure.
    show_counts:
    include cluster sizes in node labels.

    Returns:
    ---------------------
    Plotly Figure object.
    """
    # choose top_n after-clusters by size
    after_sizes = [(i, len(cl)) for i, cl in enumerate(clusters_after)]
    after_sizes.sort(key=lambda x: x[1], reverse=True)
    top_after_ids = [i for i, _ in after_sizes[:top_n]]

    # build mapping from molecule index -> after cluster id (if any)
    after_map = {}
    for after_id, cl in enumerate(clusters_after):
        for idx in cl:
            after_map[idx] = after_id

    # count flows from before_id -> after_id (only to top_after_ids)
    flow = defaultdict(int)
    before_has_flow = set()
    for before_id, cl in enumerate(clusters_before):
        for idx in cl:
            after_id = after_map.get(idx, None)
            if after_id is None:
                continue
            if after_id not in top_after_ids:
                continue
            flow[(before_id, after_id)] += 1
            before_has_flow.add(before_id)

    # prepare node lists: only before nodes that contribute to selected after nodes
    before_ids = sorted(before_has_flow)
    n_before = len(before_ids)
    # n_after = len(top_after_ids)

    # prepare labels (use 'before'/'after' prefixes)
    if show_counts:
        bef_lab = [f"before:{i} ({len(clusters_before[i])})" for i in before_ids]
        aft_lab = [f"after:{j} ({len(clusters_after[j])})" for j in top_after_ids]
    else:
        bef_lab = [f"before:{i}" for i in before_ids]
        aft_lab = [f"after:{j}" for j in top_after_ids]
    labels = bef_lab + aft_lab

    # mapping from original ids to sankey node indices
    before_index = {bid: idx for idx, bid in enumerate(before_ids)}
    after_index = {aid: n_before + i for i, aid in enumerate(top_after_ids)}

    # Build link tuples and apply min_flow and max_flow (if provided)
    links = [
        (b, a, cnt)
        for (b, a), cnt in flow.items()
        if cnt >= min_flow and (max_flow is None or cnt <= max_flow)
    ]

    # sort links by size (descending) for stable plotting
    links.sort(key=lambda x: x[2], reverse=True)

    sources = [before_index[b] for b, _, _ in links]
    targets = [after_index[a] for _, a, _ in links]
    values = [cnt for _, _, cnt in links]

    if len(values) == 0:
        print("No flows match filters; adjust min_flow/max_flow or check clusters")

    node = dict(label=labels, pad=15, thickness=20)
    link = dict(source=sources, target=targets, value=values)

    fig = go.Figure(go.Sankey(node=node, link=link))
    title_text = title or f"Cluster flow - top {len(top_after_ids)} clusters"
    fig.update_layout(
        title_text=title_text,
        font_size=10,
    )
    return fig


def plot_cluster_refinement(
    df: pd.DataFrame | None = None,
    fps: np.ndarray | None = None,
    show: bool = True,
    count_scale: float = 1e3,
    save_path: str = None,
):
    """Plot cluster refinement summary.

    Parameters
    ----------
    df : pd.DataFrame | None
        DataFrame with columns
        ['iteration',
        'n_clusters',
        'n_singletons',
        'dispersions',
        'dispersions_no_singletons'].
    fps : np.ndarray | None
        Fingerprints used to compute iSIM line. Required if df is None.
    Returns
    -------
    fig, axes
        Matplotlib Figure and Axes objects.
    """

    isim_line = iSIM.jt_isim_packed(fps)

    # Extract data
    iterations = df["iteration"].to_numpy()
    num_clusters = df["n_clusters"].to_numpy()
    num_singletons = df["n_singletons"].to_numpy()
    singleton_ratio = num_singletons / num_clusters
    dispersion_medoids = df["dispersions"].to_numpy()
    dispersion_medoids_non_singleton = df["dispersions_no_singletons"].to_numpy()

    # Create a figure with two subplots side by side
    fig, axes = plt.subplots(1, 2, figsize=(16, 6))

    # Panel 1: Clusters, Singletons, and Singleton/Cluster Ratio
    ax1 = axes[0]
    ax1.plot(iterations, num_clusters / count_scale, label="Clusters", marker="o")
    ax1.plot(iterations, num_singletons / count_scale, label="Singletons", marker="o")
    ax1.set_xlabel("Iteration", fontsize=12)
    ax1.set_ylabel(f"Count (x{int(count_scale)})", fontsize=12)
    ax1.legend(loc="upper left")

    # Twin y-axis for Singleton/Cluster Ratio
    ax2 = ax1.twinx()
    ax2.plot(
        iterations,
        singleton_ratio,
        label="Singleton/Cluster Ratio",
        color="green",
        marker="o",
    )
    ax2.set_ylabel("Singleton/Cluster Ratio", color="green", fontsize=12)
    ax2.tick_params(axis="y", labelcolor="green")

    # Panel 2: Dispersion of Medoids
    ax3 = axes[1]
    ax3.plot(iterations, dispersion_medoids, marker="o", label="All Clusters")
    ax3.plot(
        iterations,
        dispersion_medoids_non_singleton,
        marker="o",
        label="Non-Singleton Clusters",
    )
    ax3.legend(loc="upper right")
    ax3.set_xlabel("Iteration", fontsize=12)
    ax3.set_ylabel("Dispersion of Medoids (iSIM)", fontsize=12)
    ax3.hlines(
        isim_line,
        xmin=iterations[0],
        xmax=iterations[-1],
        colors="orange",
        linestyles="dashed",
    )

    plt.tight_layout()

    if save_path:
        plt.savefig(save_path, dpi=400)
    if show:
        plt.show()
    return fig, axes


def plot_threshold_scan(
    df: pd.DataFrame,
    fps: np.ndarray | None = None,
    show: bool = True,
    save_path: str = None,
    count_scale: float = 1e6,
):

    isim_line = iSIM.jt_isim_packed(fps)

    # Extract data
    thresholds = df["thresholds"].to_numpy()
    num_clusters = df["n_clusters"].to_numpy()
    num_singletons = df["n_singletons"].to_numpy()
    singleton_ratio = num_singletons / num_clusters
    dispersion_medoids = df["dispersions"].to_numpy()
    dispersion_medoids_non_singleton = df["dispersions_non_singleton"].to_numpy()

    # Create a figure with two subplots side by side
    fig, axes = plt.subplots(1, 2, figsize=(16, 6))

    # Panel 1: Clusters, Singletons, and Singleton/Cluster Ratio
    ax1 = axes[0]
    ax1.plot(thresholds, num_clusters / count_scale, label="Clusters", marker="o")
    ax1.plot(thresholds, num_singletons / count_scale, label="Singletons", marker="o")
    ax1.set_xlabel("Threshold", fontsize=12)
    ax1.set_ylabel(f"Count (x{int(count_scale)})", fontsize=12)
    ax1.set_xticks(np.arange(0, 1.1, 0.1))
    ax1.legend(loc="upper left")

    # Twin y-axis for Singleton/Cluster Ratio
    ax2 = ax1.twinx()
    ax2.plot(
        thresholds,
        singleton_ratio,
        label="Singleton/Cluster Ratio",
        color="green",
        marker="o",
    )
    ax2.set_ylabel("Singleton/Cluster Ratio", color="green", fontsize=12)
    ax2.tick_params(axis="y", labelcolor="green")

    # Add top axis for standard deviations
    ax1_top = ax1.twiny()
    n_steps = len(thresholds) - 1
    step_min, step_max = 0, n_steps
    val_min, val_max = thresholds[0], thresholds[-1]

    def threshold_to_step(x):
        return step_min + (x - val_min) * (step_max - step_min) / (val_max - val_min)

    ax1_top.set_xlim(
        threshold_to_step(ax1.get_xlim()[0]),
        threshold_to_step(ax1.get_xlim()[1]),
    )
    ax1_top.set_xticks(np.arange(0, n_steps + 1, 1))
    ax1_top.set_xlabel("Standard Deviations above iSIM", fontsize=12)

    # Panel 2: Dispersion of Medoids
    ax3 = axes[1]
    ax3.plot(thresholds, dispersion_medoids, marker="o", label="All Clusters")
    ax3.plot(
        thresholds,
        dispersion_medoids_non_singleton,
        marker="o",
        label="Non-Singleton Clusters",
    )
    ax3.legend(loc="upper right")
    ax3.set_xlabel("Threshold", fontsize=12)
    ax3.set_ylabel("Dispersion of Medoids (iSIM)", fontsize=12)
    ax3.set_xticks(np.arange(0, 1.1, 0.1))
    ax3.hlines(
        isim_line,
        xmin=thresholds[0],
        xmax=thresholds[-1],
        colors="orange",
        linestyles="dashed",
    )

    # Add top axis for standard deviations
    ax3_top = ax3.twiny()
    ax3_top.set_xlim(
        threshold_to_step(ax3.get_xlim()[0]), threshold_to_step(ax3.get_xlim()[1])
    )
    ax3_top.set_xticks(np.arange(0, n_steps + 1, 1))
    ax3_top.set_xlabel("Standard Deviations above iSIM", fontsize=12)

    # Adjust layout and show the figure
    plt.tight_layout()
    if save_path:
        plt.savefig(save_path, dpi=400)
    if show:
        plt.show()


def plot_distributions_by_fp(pairwise_similarities: dict, save_path: str = None):
    plt.figure(figsize=(8, 6))
    colors = ["blue", "orange", "green", "gray", "red"]
    for fp_type, color in zip(pairwise_similarities.keys(), colors):
        plt.hist(
            pairwise_similarities[fp_type].flatten(),
            bins=50,
            alpha=0.5,
            label=fp_type,
            color=color,
        )
        plt.axvline(
            x=np.mean(pairwise_similarities[fp_type]),
            ymin=0,
            ymax=plt.ylim()[1],
            color=color,
            linestyle="dashed",
        )
    plt.xlim(0, 1)
    plt.xticks(np.arange(0, 1.1, 0.1))
    plt.xlabel("Tanimoto Similarity")
    plt.ylabel("Frequency")
    plt.legend()
    if save_path:
        plt.savefig(save_path, dpi=400)
    else:
        plt.show()


def plot_branching_factor_scan(
    df: pd.DataFrame,
    fps: np.ndarray | None = None,
    sigmas_list: list[float] | None = None,
    save_path: str = None,
):
    df["ratio"] = df["n_singletons"] / df["n_clusters"]

    fig, axs = plt.subplots(2, 2, figsize=(12, 10))
    what_list = ["n_clusters", "n_singletons", "ratio", "dispersions_non_singleton"]

    for i, what in enumerate(what_list):
        pivot_df = df.pivot(index="branching_factor", columns="thresholds", values=what)
        ax = axs.flatten()[i]

        # Show colorbar for both, with specific range for the bottom row
        show_cbar = True
        cbar_kws = {"ticks": [0, 0.25, 0.5, 0.75, 1]} if i // 2 == 1 else {}
        im = sns.heatmap(
            pivot_df[::-1],  # reverse row order for correct y-axis orientation
            ax=ax,
            cmap="RdYlBu_r",
            vmin=0,
            vmax=1 if i // 2 == 1 else None,  # set range 0..1 for bottom row
            cbar=show_cbar,
            cbar_kws=cbar_kws,
            annot=False,
        )

        ax.set_title(what.replace("_", " ").title())

        # Only show x-axis labels on bottom row
        if i // 2 == 0:
            ax.set_xticklabels([])  # hide top-row x tick labels
            ax.set_xlabel("")
        else:
            ax.set_xlabel("Thresholds (std above iSIM)")
            ax.set_xticklabels([str(t) for t in sigmas_list], rotation=0)

        # Only show y-axis labels on left column
        if i % 2 == 1:
            ax.set_yticklabels([])  # hide right-column y tick labels
            ax.set_ylabel("")
        else:
            ax.set_ylabel("Branching Factor")

        # If this panel has a colorbar, remove its label
        if show_cbar:
            mappable = im.collections[0]
            cbar = getattr(mappable, "colorbar", None)
            if cbar is not None:
                cbar.set_label("")  # remove colorbar label
                cbar.ax.tick_params(labelsize=8)

                if what == "dispersions_non_singleton":
                    # draw marker line at min_t and annotate (clamped if outside)
                    cmin, cmax = cbar.vmin, cbar.vmax
                    y_plot = min(max(iSIM.jt_isim_packed(fps), cmin), cmax)
                    # draw horizontal line (data coords)
                    cbar.ax.axhline(y_plot, color="k", linewidth=2, zorder=10)

    plt.tight_layout()
    if save_path:
        plt.savefig(save_path, dpi=400)
    else:
        plt.show()


def sampled_MSC(cluster, smiles, n_samples=100):
    """Calculate the MSC for a sampled subset of molecules in a cluster."""
    n_samples = min(n_samples, len(cluster))

    # Get the SMILES for the sampled molecules
    remover = SaltRemover.SaltRemover()
    sampled_smiles = [smiles[i] for i in cluster[:n_samples]]
    mols = [Chem.MolFromSmiles(smi) for smi in sampled_smiles]
    mols = [remover.StripMol(mol) for mol in mols]

    # Find the MSC of those molecules
    MSC = rdFMCS.FindMCS(mols, threshold=0.75)
    MSC_mol = Chem.MolFromSmarts(MSC.smartsString)
    img2 = Draw.MolToImage(MSC_mol)

    # Find substructure matches in the sampled molecules with the MSC
    for mol in mols:
        if mol.HasSubstructMatch(MSC_mol):
            match = mol.GetSubstructMatch(MSC_mol)
            atom_indices = list(match)
            highlight_atoms = atom_indices
            # Highlight the matching substructure
            mol.SetProp("_highlightAtoms", ",".join(map(str, highlight_atoms)))

    highlight_lists = []
    for mol in mols:
        if mol.HasProp("_highlightAtoms"):
            vals = mol.GetProp("_highlightAtoms").split(",")
            highlight_lists.append(list(map(int, vals)))
        else:
            highlight_lists.append([])
    img1 = Draw.MolsToGridImage(
        mols, highlightAtomLists=highlight_lists, molsPerRow=5, subImgSize=(200, 200)
    )
    return img1, img2



r"""Pretty printing"""

from pathlib import Path
import numpy as np
import typing as tp
import os
import multiprocessing as mp

from rich.console import Console

from bblean._memory import get_peak_memory_gib


class BBConsole(Console):
    def print_banner(self) -> None:
        if os.environ.get("BITBIRCHNOBANNER", ""):
            return
        banner = r"""[bold]
        ______ _ _  ______ _          _        
        | ___ (_) | | ___ (_)        | |     [/bold][cyan]  ______                      [/cyan][bold]
        | |_/ /_| |_| |_/ /_ _ __ ___| |__   [/bold][cyan]  ___  / ___________ _______  [/cyan][bold]
        | ___ \ | __| ___ \ | '__/ __| '_ \  [/bold][cyan]  __  /  _  _ \  __ `/_  __ \ [/cyan][bold]
        | |_/ / | |_| |_/ / | | | (__| | | | [/bold][cyan]  _  /___/  __/ /_/ /_  / / / [/cyan][bold]
        \____/|_|\__\____/|_|_|  \___|_| |_| [/bold][cyan]  /_____/\___/\__,_/ /_/ /_/  [/cyan][bold]"""  # noqa W291
        self.print(banner, highlight=False)
        self.print()
        self.print()
        self.print(
            r"""BitBirch-Lean is developed by the [bold]Miranda-Quintana Lab[/bold] https://github.com/mqcomplab
If you find this software useful please cite the following articles:
    [yellow]•[/yellow] [italic]BitBIRCH: efficient clustering of large molecular libraries[/italic]:
        https://doi.org/10.1039/D5DD00030K
    [yellow]•[/yellow] [italic]BitBIRCH Clustering Refinement Strategies[/italic]:
        https://doi.org/10.1021/acs.jcim.5c00627
    [yellow]•[/yellow] [italic]BitBIRCH-Lean[/italic]:
        (preprint) https://www.biorxiv.org/content/10.1101/2025.10.22.684015v1"""  # noqa
        )

    def print_peak_mem(self, out_dir: Path, indent: bool = True) -> None:
        peak_mem_gib = get_peak_memory_gib(out_dir)
        if peak_mem_gib is None:
            return
        indent_str = " " * 4 if indent else ""
        self.print(f"{indent_str}- Peak RAM use: {peak_mem_gib:.4f} GiB")

    def print_config(self, config: dict[str, tp.Any]) -> None:
        num_fps_loaded = np.array(config["num_fps_loaded"])
        total_fps_num = num_fps_loaded.sum()
        with np.printoptions(formatter={"int": "{:,}".format}, threshold=10):
            num_fps_str = str(num_fps_loaded)[1:-1]
            self.print(
                f"Running [bold]single-round, serial (1 process)[/bold] clustering\n\n"
                f"- Branching factor: {config['branching_factor']:,}\n"
                f"- Merge criterion: [yellow]{config['merge_criterion']}[/yellow]\n"
                f"- Threshold: {config['threshold']}\n"
                f"- Num. files loaded: {len(config['input_files']):,}\n"
                f"- Num. fingerprints loaded for each file: {num_fps_str}\n"
                f"- Total num. fingerprints: {total_fps_num:,}\n"
                f"- Output directory: {config['out_dir']}\n",
                end="",
            )
        bb_variant = config.get("bitbirch_variant", "lean")
        max_files = config.get("max_files", None)
        max_fps = config.get("max_fps", None)
        if "tolerance" in config["merge_criterion"]:
            self.print(f"- Tolerance: {config['tolerance']}\n", end="")
        if config["refine_num"] > 0:
            self.print(
                f"- Will refine largest {config['refine_num']} clusters\n", end=""
            )
            self.print(f"- Num. clusters to refine: {config['refine_num']}\n", end="")
            self.print(
                "- Refine criterion: "
                f"[yellow]{config['refine_merge_criterion']}[/yellow]\n",
                end="",
            )
            if "tolerance" in config["refine_merge_criterion"]:
                self.print(f"- Refine tolerance: {config['tolerance']}\n", end="")
            self.print(
                f"- Refine threshold change: {config['refine_threshold_change']}\n",
                end="",
            )
        if bb_variant != "lean":
            self.print(
                "- [bold]DEBUG:[/bold] Using bitbirch version: {variant}\n", end=""
            )
        if max_files is not None:
            self.print(
                f"- [bold]DEBUG:[/bold] Max files to load: {max_files:,}\n", end=""
            )
        if max_fps is not None:
            self.print(
                f"- [bold]DEBUG:[/bold] Max fps to load per file: {max_fps:,}\n", end=""
            )
        self.print()

    def print_multiround_config(
        self, config: dict[str, tp.Any], mp_context: tp.Any = None
    ) -> None:
        if mp_context is None:
            mp_context = mp.get_context()
        num_processes = config.get("num_initial_processes", 1)
        extra_desc = (
            f"parallel (max {num_processes:,} processes)"
            if num_processes > 1
            else "serial (1 process)"
        )
        desc = f"multi-round, {extra_desc}"
        num_fps_loaded = np.array(config["num_fps_loaded"])
        total_fps_num = num_fps_loaded.sum()
        with np.printoptions(formatter={"int": "{:,}".format}, threshold=10):
            num_fps_str = str(num_fps_loaded)[1:-1]
            self.print(
                f"Running [bold]{desc}[/bold] clustering\n\n"
                f"- Branching factor: {config['branching_factor']:,}\n"
                f"- Initial round merge criterion: [yellow]{config['initial_merge_criterion']}[/yellow]\n"  # noqa:E501
                f"- Threshold: {config['threshold']}\n"
                f"- Tolerance: {config['tolerance']}\n"
                f"- Num. files loaded: {len(config['input_files']):,}\n"
                f"- Num. fingerprints loaded for each file: {num_fps_str}\n"
                f"- Total num. fingerprints: {total_fps_num:,}\n"
                f"- Output directory: {config['out_dir']}\n",
                end="",
            )
        full_refinement_before_midsection = config.get(
            "full_refinement_before_midsection", False
        )
        bb_variant = config.get("bitbirch_variant", "lean")
        max_files = config.get("max_files", None)
        bin_size = config.get("bin_size", None)
        max_fps = config.get("max_fps", None)
        if bin_size is not None:
            self.print(f"- Bin size for second round: {bin_size:,}\n", end="")
        if num_processes > 1:
            self.print(
                f"- Multiprocessing method: [yellow]{mp_context._name}[/yellow]\n",
                end="",
            )
        if not full_refinement_before_midsection:
            self.print(
                f"- Full refinement before midsection: {full_refinement_before_midsection}\n",  # noqa:E501
                end="",
            )
        if bb_variant != "lean":
            self.print(
                "- [bold]DEBUG:[/bold] Using bitbirch version: {variant}\n", end=""
            )
        if max_files is not None:
            self.print(
                f"- [bold]DEBUG:[/bold] Max files to load: {max_files:,}\n", end=""
            )
        if max_fps is not None:
            self.print(
                f"- [bold]DEBUG:[/bold] Max fps to load per file: {max_fps:,}\n", end=""
            )
        self.print()


class SilentConsole(BBConsole):
    def print(self, *args: tp.Any, **kwargs: tp.Any) -> None:
        pass

    def print_peak_mem(self, out_dir: Path, indent: bool = True) -> None:
        pass

    def print_banner(self) -> None:
        pass

    def status(self, *args: tp.Any, **kwargs: tp.Any) -> tp.Any:
        class DummyStatus:
            def __enter__(self) -> tp.Any:
                return self

            def __exit__(self, *args: tp.Any, **kwargs: tp.Any) -> None:
                pass

        return DummyStatus()


_console = BBConsole()
_silent_console = SilentConsole()


def get_console(silent: bool = False) -> BBConsole:
    if silent:
        return _silent_console
    return _console



import numpy as np

from bblean.bitbirch import BitBirch
from bblean.fingerprints import make_fake_fingerprints, unpack_fingerprints
from inline_snapshot import snapshot


def test_random_fps_consistency() -> None:
    # TODO For some strange reason this test *is non deterministic*
    # The issue seems to be much workes in macOS and Windows than in linux, but
    # it happens in all cases
    # The kmeans implementation of sklearn seems to work different in linux and macOS
    fps = make_fake_fingerprints(3000, n_features=2048, seed=126205095409235, pack=True)
    tree = BitBirch(branching_factor=50, threshold=0.65, merge_criterion="diameter")
    tree.fit(fps, n_features=2048)
    output_cent = tree.get_centroids()
    output_med = tree.get_medoids(fps)
    assert [c.tolist()[:5] for c in output_cent[:5]] == snapshot(
        [
            [255, 255, 255, 255, 255],
            [255, 251, 255, 255, 255],
            [255, 239, 255, 247, 255],
            [255, 255, 191, 255, 255],
            [255, 127, 235, 127, 255],
        ]
    )
    assert output_med[:5, :5].tolist() == snapshot(
        [
            [255, 255, 126, 255, 111],
            [247, 255, 255, 255, 255],
            [191, 253, 191, 255, 255],
            [255, 255, 95, 255, 239],
            [235, 255, 123, 255, 255],
        ]
    )

    tree.global_clustering(
        20,
        method="kmeans",
        n_init=1,
        init=unpack_fingerprints(np.vstack(output_cent))[::2][:20],
        max_iter=10,
    )
    output_mol_ids = tree.get_cluster_mol_ids(global_clusters=True, sort=False)
    output_med = tree.get_medoids(fps, global_clusters=True, sort=False)
    assert output_mol_ids
    assert output_med.any()



import itertools
import pytest
import numpy as np

from bblean.bitbirch import BitBirch, guess_threshold
from bblean.fingerprints import pack_fingerprints, make_fake_fingerprints

from inline_snapshot import snapshot

# NOTE: Results on this file don't depend on branching factor / threshold


def test_bb_lean_defaults() -> None:
    tree = BitBirch()
    assert tree.branching_factor == 50
    assert tree.threshold == 0.65
    assert tree.merge_criterion == "diameter"


def test_bb_cluster_empty_input() -> None:
    fp = pack_fingerprints(np.zeros((0, 2048), dtype=np.uint8))
    # At least 1 fp should be present
    with pytest.raises(ValueError):
        _ = BitBirch().fit(fp, n_features=2048).get_cluster_mol_ids()


def test_bb_cluster_simple_repeated_fps() -> None:
    for repeats in (1, 2, 10):
        zeros_fp = pack_fingerprints(np.zeros((repeats, 2048), dtype=np.uint8))
        ids = BitBirch().fit(zeros_fp, n_features=2048).get_cluster_mol_ids()
        assert ids == [list(range(repeats))]

        ones_fp = pack_fingerprints(np.ones((repeats, 2048), dtype=np.uint8))
        ids = BitBirch().fit(ones_fp, n_features=2048).get_cluster_mol_ids()
        assert ids == [list(range(repeats))]

        rng = np.random.default_rng(12620509540149709235)
        mixed_fp = pack_fingerprints(
            np.tile(rng.integers(0, 2, (1, 2048), dtype=np.uint8), (repeats, 1))
        )
        ids = BitBirch().fit(mixed_fp, n_features=2048).get_cluster_mol_ids()
        assert ids == [list(range(repeats))]


def test_guess_threhsold() -> None:
    fps = make_fake_fingerprints(
        100, n_features=8, seed=12620509540149709235, pack=True
    )
    thresh = guess_threshold(fps, return_mean_std=False)
    assert thresh > 0.9 and thresh < 1.0

    fps = make_fake_fingerprints(
        100, n_features=2048, seed=12620509540149709235, pack=True
    )
    thresh = guess_threshold(fps, return_mean_std=False)
    assert thresh > 0.4 and thresh < 0.6


def test_bb_cluster_3_fps() -> None:
    fps = make_fake_fingerprints(3, n_features=8, seed=12620509540149709235, pack=True)

    data = BitBirch().fit(fps).get_cluster_mol_ids()
    assert data == snapshot([[0], [1], [2]])
    data = BitBirch().fit(fps, weights=itertools.repeat(5)).get_cluster_mol_ids()
    assert data == snapshot([[1, 2], [0]])
    data = BitBirch().fit(fps, weights=itertools.repeat(10000)).get_cluster_mol_ids()
    assert data == snapshot([[1, 2], [0]])
    data = BitBirch().fit(fps, weights=itertools.repeat(1000000)).get_cluster_mol_ids()
    assert data == snapshot([[1, 2], [0]])



r"""Clustering metrics using Tanimoto similarity"""

from contextlib import nullcontext

import numpy as np
from numpy.typing import NDArray
from rich.progress import Progress

from bblean.similarity import (
    jt_isim_from_sum,
    jt_sim_packed,
    jt_isim_packed,
    jt_isim_unpacked,
    centroid as centroid_from_fps,
    centroid_from_sum,
    jt_isim_medoid,
)
from bblean.fingerprints import unpack_fingerprints, pack_fingerprints

__all__ = ["jt_isim_chi", "jt_isim_dunn", "jt_dbi"]


def _calc_centrals(
    cluster_fps: list[NDArray[np.uint8]],
    kind: str,
    input_is_packed: bool = True,
    n_features: int | None = None,
    pack: bool = True,
) -> list[NDArray[np.uint8]]:
    if kind == "medoid":
        return [
            jt_isim_medoid(
                c, input_is_packed=input_is_packed, n_features=n_features, pack=pack
            )[1]
            for c in cluster_fps
        ]
    elif kind == "centroid":
        return [
            centroid_from_fps(
                c, input_is_packed=input_is_packed, n_features=n_features, pack=pack
            )
            for c in cluster_fps
        ]
    raise ValueError(f"Unknown arg {kind} use 'medoids|centroids'")


def jt_isim_chi(
    cluster_fps: list[NDArray[np.uint8]],
    all_fps_central: NDArray[np.uint8] | str = "centroid",
    centrals: list[NDArray[np.uint8]] | str = "centroid",
    input_is_packed: bool = True,
    n_features: int | None = None,
    verbose: bool = False,
) -> float:
    """Calinski-Harabasz clustering index

    An approximation to the CHI index using the Tanimoto iSIM. *Higher* is better.
    """
    all_fps_num = sum(len(c) for c in cluster_fps)
    if isinstance(all_fps_central, str):
        if not all_fps_central == "centroid":
            raise NotImplementedError("Currently only 'centroid' implemented for CHI")
        if input_is_packed:
            unpacked_clusts = [unpack_fingerprints(c, n_features) for c in cluster_fps]
        else:
            unpacked_clusts = cluster_fps
        total_linear_sum = sum(np.sum(c, axis=0) for c in unpacked_clusts)
        all_fps_central = centroid_from_sum(total_linear_sum, all_fps_num)

    if isinstance(centrals, str):
        if not centrals == "centroid":
            raise NotImplementedError("Currently only 'centroid' implemented for CHI")
        centrals = _calc_centrals(cluster_fps, centrals, input_is_packed, n_features)
    else:
        if not input_is_packed:
            centrals = [pack_fingerprints(c) for c in centrals]

    clusters_num = len(cluster_fps)
    # Packed cluster_fps required for CHI
    if not input_is_packed:
        cluster_fps = [pack_fingerprints(c) for c in cluster_fps]

    if clusters_num <= 1:
        return 0

    wcss = 0.0  # within-cluster sum of squares
    bcss = 0.0  # between-cluster sum of squares
    progress = Progress(transient=True) if verbose else nullcontext()
    with progress as pbar:
        if verbose:
            task = pbar.add_task(  # type: ignore
                "[italic]Calculating CHI[/italic]...",
                total=(len(centrals)),
            )
        for central, clust in zip(centrals, cluster_fps):
            # NOTE: In the original implementation there isn't a (1 - jt...) here (!)
            bcss += (
                len(clust) * (1 - jt_sim_packed(all_fps_central, central).item()) ** 2
            )
            d = 1 - jt_sim_packed(clust, central)
            wcss += np.dot(d, d)
            if verbose:
                pbar.update(task, advance=1)  # type: ignore
    # TODO: When can the denom be 0?
    return bcss * (all_fps_num - clusters_num) / (wcss * (clusters_num - 1))


def jt_dbi(
    cluster_fps: list[NDArray[np.uint8]],
    centrals: list[NDArray[np.uint8]] | str = "centroid",
    input_is_packed: bool = True,
    n_features: int | None = None,
    verbose: bool = False,
) -> float:
    """Davies-Bouldin clustering index

    DBI index using the Tanimoto distance. *Lower* is better.
    """
    if isinstance(centrals, str):
        centrals = _calc_centrals(cluster_fps, centrals, input_is_packed, n_features)
    else:
        if not input_is_packed:
            centrals = [pack_fingerprints(c) for c in centrals]

    # Centrals can be 'medoids' or 'centroids'
    if not input_is_packed:
        cluster_fps = [pack_fingerprints(c) for c in cluster_fps]
    # Packed cluster_fps required for DBI

    fps_num = 0
    S: list[float] = []
    for central, clust_fps in zip(centrals, cluster_fps):
        size = len(clust_fps)
        S.append(np.sum(1 - jt_sim_packed(clust_fps, central)) / size)
        fps_num += size

    if fps_num == 0:
        return 0

    # Quadratic scaling on num. clusters
    progress = Progress(transient=True) if verbose else nullcontext()
    with progress as pbar:
        if verbose:
            task = pbar.add_task(  # type: ignore
                "[italic]Calculating DBI[/italic]...",
                total=(len(centrals) ** 2 - len(centrals)),
            )
        numerator = 0.0
        for i, central in enumerate(centrals):
            max_d = 0.0
            for j, other_central in enumerate(centrals):
                if i == j:
                    continue
                Mij = 1 - jt_sim_packed(central, other_central).item()
                max_d = max(max_d, (S[i] + S[j]) / Mij)
                if verbose:
                    pbar.update(task, advance=1)  # type: ignore
            numerator += max_d
    return numerator / fps_num


# This is the Dunn varaint used in the original BitBirch article
def jt_isim_dunn(
    cluster_fps: list[NDArray[np.uint8]],
    input_is_packed: bool = True,
    n_features: int | None = None,
    verbose: bool = False,
) -> float:
    """Dunn clustering index

    An approximation to the Dunn index using the Tanimoto iSIM. *Higher* is better.
    """
    # Unpacked cluster_fps required for Dunn
    if input_is_packed:
        D = [jt_isim_packed(clust) for clust in cluster_fps]
        cluster_fps = [unpack_fingerprints(clust, n_features) for clust in cluster_fps]
    else:
        D = [jt_isim_unpacked(clust) for clust in cluster_fps]
    max_d = max(D)
    if max_d == 0:
        # TODO: Unclear what to return in this case, probably 1.0 is safer?
        return 1
    min_d = 1.00
    # Quadratic scaling on num. clusters
    pairs_num = len(cluster_fps) * (len(cluster_fps) - 1) // 2
    progress = Progress(transient=True) if verbose else nullcontext()
    with progress as pbar:
        if verbose:
            task = pbar.add_task(  # type: ignore
                "[italic]Calculating Dunn (slow)[/italic]...", total=pairs_num
            )
        for i, clust1 in enumerate(cluster_fps[:-1]):
            for j, clust2 in enumerate(cluster_fps[i + 1 :]):
                combined = np.sum(clust1, axis=0) + np.sum(clust2, axis=0)
                dij = 1 - jt_isim_from_sum(combined, len(clust1) + len(clust2))
                min_d = min(dij, min_d)
                if verbose:
                    pbar.update(task, advance=1)  # type: ignore
    return min_d / max(D)
